#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2018, Ansible by Red Hat, inc
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function


__metaclass__ = type


DOCUMENTATION = """
module: cli_config
author: Trishna Guha (@trishnaguha)
notes:
- The commands will be returned only for platforms that do not support onbox diff.
  The C(--diff) option with the playbook will return the difference in configuration
  for devices that has support for onbox diff
- To ensure idempotency and correct diff the configuration lines in the relevant module
  options should be similar to how they appear if present in the running configuration on
  device including the indentation.
short_description: Push text based configuration to network devices over network_cli
description:
- This module provides platform agnostic way of pushing text based configuration to
  network devices over network_cli connection plugin.
version_added: 1.0.0
extends_documentation_fragment:
- ansible.netcommon.network_agnostic
options:
  config:
    description:
    - The config to be pushed to the network device. This argument is mutually exclusive
      with C(rollback) and either one of the option should be given as input. To ensure
      idempotency and correct diff the configuration lines should be similar to how they
      appear if present in the running configuration on device including the indentation.
    type: str
  commit:
    description:
    - The C(commit) argument instructs the module to push the configuration to the
      device. This is mapped to module check mode.
    type: bool
  replace:
    description:
    - If the C(replace) argument is set to C(yes), it will replace the entire running-config
      of the device with the C(config) argument value. For devices that support replacing
      running configuration from file on device like NXOS/JUNOS, the C(replace) argument
      takes path to the file on the device that will be used for replacing the entire
      running-config. The value of C(config) option should be I(None) for such devices.
      Nexus 9K devices only support replace. Use I(net_put) or I(nxos_file_copy) in
      case of NXOS module to copy the flat file to remote device and then use set
      the fullpath to this argument.
    type: str
  backup:
    description:
    - This argument will cause the module to create a full backup of the current running
      config from the remote device before any changes are made. If the C(backup_options)
      value is not given, the backup file is written to the C(backup) folder in the
      playbook root directory or role root directory, if playbook is part of an ansible
      role. If the directory does not exist, it is created.
    type: bool
    default: no
  rollback:
    description:
    - The C(rollback) argument instructs the module to rollback the current configuration
      to the identifier specified in the argument.  If the specified rollback identifier
      does not exist on the remote device, the module will fail. To rollback to the
      most recent commit, set the C(rollback) argument to 0. This option is mutually
      exclusive with C(config).
    type: int
  commit_comment:
    description:
    - The C(commit_comment) argument specifies a text string to be used when committing
      the configuration. If the C(commit) argument is set to False, this argument
      is silently ignored. This argument is only valid for the platforms that support
      commit operation with comment.
    type: str
  defaults:
    description:
    - The I(defaults) argument will influence how the running-config is collected
      from the device.  When the value is set to true, the command used to collect
      the running-config is append with the all keyword.  When the value is set to
      false, the command is issued without the all keyword.
    default: no
    type: bool
  multiline_delimiter:
    description:
    - This argument is used when pushing a multiline configuration element to the
      device. It specifies the character to use as the delimiting character. This
      only applies to the configuration action.
    type: str
  diff_replace:
    description:
    - Instructs the module on the way to perform the configuration on the device.
      If the C(diff_replace) argument is set to I(line) then the modified lines are
      pushed to the device in configuration mode. If the argument is set to I(block)
      then the entire command block is pushed to the device in configuration mode
      if any line is not correct. Note that this parameter will be ignored if the
      platform has onbox diff support.
    type: str
    choices:
    - line
    - block
    - config
  diff_match:
    description:
    - Instructs the module on the way to perform the matching of the set of commands
      against the current device config. If C(diff_match) is set to I(line), commands
      are matched line by line. If C(diff_match) is set to I(strict), command lines
      are matched with respect to position. If C(diff_match) is set to I(exact), command
      lines must be an equal match. Finally, if C(diff_match) is set to I(none), the
      module will not attempt to compare the source configuration with the running
      configuration on the remote device. Note that this parameter will be ignored
      if the platform has onbox diff support.
    type: str
    choices:
    - line
    - strict
    - exact
    - none
  diff_ignore_lines:
    description:
    - Use this argument to specify one or more lines that should be ignored during
      the diff. This is used for lines in the configuration that are automatically
      updated by the system. This argument takes a list of regular expressions or
      exact line matches. Note that this parameter will be ignored if the platform
      has onbox diff support.
    type: list
    elements: str
  backup_options:
    description:
    - This is a dict object containing configurable options related to backup file
      path. The value of this option is read only when C(backup) is set to I(yes),
      if C(backup) is set to I(no) this option will be silently ignored.
    suboptions:
      filename:
        description:
        - The filename to be used to store the backup configuration. If the filename
          is not given it will be generated based on the hostname, current time and
          date in format defined by <hostname>_config.<current-date>@<current-time>
        type: str
      dir_path:
        description:
        - This option provides the path ending with directory name in which the backup
          configuration file will be stored. If the directory does not exist it will
          be first created and the filename is either the value of C(filename) or
          default filename as described in C(filename) options description. If the
          path value is not given in that case a I(backup) directory will be created
          in the current working directory and backup configuration will be copied
          in C(filename) within I(backup) directory.
        type: path
    type: dict
"""

EXAMPLES = """
- name: configure device with config
  ansible.netcommon.cli_config:
    config: "{{ lookup('ansible.builtin.template', 'basic/config.j2') }}"

- name: multiline config
  ansible.netcommon.cli_config:
    config: |
      hostname foo
      feature nxapi

- name: configure device with config with defaults enabled
  ansible.netcommon.cli_config:
    config: "{{ lookup('ansible.builtin.template', 'basic/config.j2') }}"
    defaults: "yes"

- name: Use diff_match
  ansible.netcommon.cli_config:
    config: "{{ lookup('file', 'interface_config') }}"
    diff_match: none

- name: nxos replace config
  ansible.netcommon.cli_config:
    replace: bootflash:nxoscfg

- name: junos replace config
  ansible.netcommon.cli_config:
    replace: /var/home/ansible/junos01.cfg

- name: commit with comment
  ansible.netcommon.cli_config:
    config: set system host-name foo
    commit_comment: this is a test

- name: configurable backup path
  ansible.netcommon.cli_config:
    config: "{{ lookup('ansible.builtin.template', 'basic/config.j2') }}"
    backup: "yes"
    backup_options:
      filename: backup.cfg
      dir_path: /home/user
"""

RETURN = """
diff:
  description: The diff generated on the device when the commands were applied
  returned: When I(supports_onbox_diff=True) in the platform's cliconf plugin
  type: str
  sample: |-
    --- system:/running-config
    +++ session:/ansible_1599745461-session-config
    @@ -4,7 +4,7 @@
     !
     transceiver qsfp default-mode 4x10G
     !
    -hostname veos
    +hostname veos3
     !
     spanning-tree mode mstp
commands:
  description: The set of commands that will be pushed to the remote device
  returned: When I(supports_generated_diff=True) and I(supports_onbox_diff=False) in the platform's cliconf plugin
  type: list
  sample: ['interface Loopback999', 'no shutdown']
backup_path:
  description: The full path to the backup file
  returned: when backup is yes
  type: str
  sample: /playbooks/ansible/backup/hostname_config.2016-07-16@22:28:34
"""

import json

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_text
from ansible.module_utils.connection import Connection


def validate_args(module, device_operations):
    """validate param if it is supported on the platform"""
    feature_list = [
        "replace",
        "rollback",
        "commit_comment",
        "defaults",
        "multiline_delimiter",
        "diff_replace",
        "diff_match",
        "diff_ignore_lines",
    ]

    for feature in feature_list:
        if module.params[feature]:
            supports_feature = device_operations.get("supports_%s" % feature)
            if supports_feature is None:
                module.fail_json(
                    msg="This platform does not specify whether %s is supported or not. "
                    "Please report an issue against this platform's cliconf plugin." % feature
                )
            elif not supports_feature:
                module.fail_json(msg="Option %s is not supported on this platform" % feature)


def run(module, device_operations, connection, candidate, running, rollback_id):
    result = {}
    resp = {}
    config_diff = []
    banner_diff = {}

    replace = module.params["replace"]
    commit_comment = module.params["commit_comment"]
    multiline_delimiter = module.params["multiline_delimiter"]
    diff_replace = module.params["diff_replace"]
    diff_match = module.params["diff_match"]
    diff_ignore_lines = module.params["diff_ignore_lines"]

    commit = not module.check_mode

    if replace in ("yes", "true", "True"):
        replace = True
    elif replace in ("no", "false", "False"):
        replace = False

    if replace is not None and replace not in [True, False] and candidate is not None:
        module.fail_json(
            msg="Replace value '%s' is a configuration file path already"
            " present on the device. Hence 'replace' and 'config' options"
            " are mutually exclusive" % replace
        )

    if rollback_id is not None:
        resp = connection.rollback(rollback_id, commit)
        if "diff" in resp:
            result["changed"] = True

    elif device_operations.get("supports_onbox_diff"):
        if diff_replace:
            module.warn("diff_replace is ignored as the device supports onbox diff")
        if diff_match:
            module.warn("diff_match is ignored as the device supports onbox diff")
        if diff_ignore_lines:
            module.warn("diff_ignore_lines is ignored as the device supports onbox diff")

        if candidate and not isinstance(candidate, list):
            candidate = candidate.strip("\n").splitlines()

        kwargs = {
            "candidate": candidate,
            "commit": commit,
            "replace": replace,
            "comment": commit_comment,
        }
        resp = connection.edit_config(**kwargs)

        if "diff" in resp:
            result["changed"] = True
            result["diff"] = resp["diff"]

    elif device_operations.get("supports_generate_diff"):
        kwargs = {"candidate": candidate, "running": running}
        if diff_match:
            kwargs.update({"diff_match": diff_match})
        if diff_replace:
            kwargs.update({"diff_replace": diff_replace})
        if diff_ignore_lines:
            kwargs.update({"diff_ignore_lines": diff_ignore_lines})

        diff_response = connection.get_diff(**kwargs)

        config_diff = diff_response.get("config_diff")
        banner_diff = diff_response.get("banner_diff")

        if config_diff:
            if isinstance(config_diff, list):
                candidate = config_diff
            else:
                candidate = config_diff.splitlines()

            kwargs = {
                "candidate": candidate,
                "commit": commit,
                "replace": replace,
                "comment": commit_comment,
            }
            if commit:
                connection.edit_config(**kwargs)
            result["changed"] = True
            result["commands"] = config_diff.split("\n")

        if banner_diff:
            candidate = json.dumps(banner_diff)

            kwargs = {"candidate": candidate, "commit": commit}
            if multiline_delimiter:
                kwargs.update({"multiline_delimiter": multiline_delimiter})
            if commit:
                connection.edit_banner(**kwargs)
            result["changed"] = True

    if module._diff:
        if "diff" in resp:
            result["diff"] = {"prepared": resp["diff"]}
        else:
            diff = ""
            if config_diff:
                if isinstance(config_diff, list):
                    diff += "\n".join(config_diff)
                else:
                    diff += config_diff
            if banner_diff:
                diff += json.dumps(banner_diff)
            result["diff"] = {"prepared": diff}

    if result.get("changed"):
        msg = (
            "To ensure idempotency and correct diff the input configuration lines should be"
            " similar to how they appear if present in"
            " the running configuration on device including the indentation"
        )
        if "warnings" in result:
            result["warnings"].append(msg)
        else:
            result["warnings"] = msg
    return result


def main():
    """main entry point for execution"""
    backup_spec = dict(filename=dict(), dir_path=dict(type="path"))
    argument_spec = dict(
        backup=dict(default=False, type="bool"),
        backup_options=dict(type="dict", options=backup_spec),
        config=dict(type="str"),
        commit=dict(type="bool"),
        replace=dict(type="str"),
        rollback=dict(type="int"),
        commit_comment=dict(type="str"),
        defaults=dict(default=False, type="bool"),
        multiline_delimiter=dict(type="str"),
        diff_replace=dict(choices=["line", "block", "config"]),
        diff_match=dict(choices=["line", "strict", "exact", "none"]),
        diff_ignore_lines=dict(type="list", elements="str"),
    )

    mutually_exclusive = [("config", "rollback")]
    required_one_of = [["backup", "config", "rollback"]]

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=mutually_exclusive,
        required_one_of=required_one_of,
        supports_check_mode=True,
    )

    result = {"changed": False}

    connection = Connection(module._socket_path)
    capabilities = module.from_json(connection.get_capabilities())

    if capabilities:
        device_operations = capabilities.get("device_operations", dict())
        validate_args(module, device_operations)
    else:
        device_operations = dict()

    if module.params["defaults"]:
        if "get_default_flag" in capabilities.get("rpc"):
            flags = connection.get_default_flag()
        else:
            flags = "all"
    else:
        flags = []

    candidate = module.params["config"]
    candidate = to_text(candidate, errors="surrogate_then_replace") if candidate else None
    running = connection.get_config(flags=flags)
    rollback_id = module.params["rollback"]

    if module.params["backup"]:
        result["__backup__"] = running

    if candidate or rollback_id is not None or module.params["replace"]:
        try:
            result.update(
                run(
                    module,
                    device_operations,
                    connection,
                    candidate,
                    running,
                    rollback_id,
                )
            )
        except Exception as exc:
            module.fail_json(msg=to_text(exc))

    module.exit_json(**result)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
Utility functions for Intersight vNIC-related modules.

This module contains common functions used by both intersight_lan_connectivity_policy
and intersight_vnic_template modules to prevent code duplication.
"""

from __future__ import absolute_import, division, print_function
__metaclass__ = type


def get_policy_moid_with_org(intersight, policy_cache, module, resource_path, policy_name, organization_name, policy_type="Policy"):
    """
    Get policy MOID with caching and organization scoping to avoid redundant API calls.

    This function scopes the policy lookup to a specific organization, preventing
    conflicts when multiple organizations have policies with the same name.

    Args:
        intersight: IntersightModule instance
        policy_cache: Dictionary for caching policy MOIDs
        module: AnsibleModule instance
        resource_path: API resource path for the policy
        policy_name: Name of the policy to resolve
        organization_name: Name of the organization to scope the search to
        policy_type: Type of policy for error messages (default: "Policy")

    Returns:
        MOID of the policy or fails with error message if not found
    """

    if not policy_name:
        return None

    cache_key = f"{resource_path}:{policy_name}"
    if cache_key in policy_cache:
        return policy_cache[cache_key]

    policy_moid = intersight.get_moid_by_name_and_org(
        resource_path=resource_path,
        resource_name=policy_name,
        organization_name=organization_name
    )
    if not policy_moid:
        module.fail_json(msg=f"{policy_type} '{policy_name}' not found in organization '{organization_name}'")

    policy_cache[cache_key] = policy_moid
    return policy_moid


def validate_cdn_config(module, vnic_config):
    """
    Validate CDN (Consistent Device Naming) configuration.
    """
    cdn_source = vnic_config.get('cdn_source', 'vnic')
    cdn_value = vnic_config.get('cdn_value')
    vnic_name = vnic_config.get('name', 'unknown')

    if cdn_source == 'user' and not cdn_value:
        module.fail_json(msg=f"cdn_value is required when cdn_source is set to 'user' for {vnic_name}")

    if cdn_source == 'vnic' and cdn_value:
        module.fail_json(msg=f"cdn_value should not be provided when cdn_source is set to 'vnic' for {vnic_name}")


def validate_usnic_settings(module, usnic_settings, name=None):
    """
    Validate USNIC connection type settings.
    """
    error_suffix = f" for {name}" if name else ""

    if not usnic_settings:
        module.fail_json(msg=f"usnic_settings is required when connection_type is 'usnic'{error_suffix}")

    usnic_adapter_policy_name = usnic_settings.get('usnic_adapter_policy_name')
    if not usnic_adapter_policy_name:
        module.fail_json(msg=f"usnic_adapter_policy_name is required in usnic_settings{error_suffix}")

    count = usnic_settings.get('count', 0)
    cos = usnic_settings.get('cos', 5)

    if count < 0 or count > 225:
        module.fail_json(msg=f"USNIC count must be between 0 and 225{error_suffix}")
    if cos < 0 or cos > 6:
        module.fail_json(msg=f"USNIC CoS must be between 0 and 6{error_suffix}")


def validate_vmq_settings(module, vmq_settings, name=None):
    """
    Validate VMQ connection type settings.
    """
    error_suffix = f" for {name}" if name else ""
    multi_queue_support = vmq_settings.get('multi_queue_support', False)

    if not multi_queue_support:
        num_interrupts = vmq_settings.get('num_interrupts', 16)
        num_vmqs = vmq_settings.get('num_vmqs', 4)

        if num_interrupts < 1 or num_interrupts > 514:
            module.fail_json(msg=f"VMQ num_interrupts must be between 1 and 514{error_suffix}")
        if num_vmqs < 1 or num_vmqs > 128:
            module.fail_json(msg=f"VMQ num_vmqs must be between 1 and 128{error_suffix}")
    else:
        num_sub_vnics = vmq_settings.get('num_sub_vnics', 64)
        if num_sub_vnics < 0 or num_sub_vnics > 64:
            module.fail_json(msg=f"VMQ num_sub_vnics must be between 0 and 64{error_suffix}")

        if not vmq_settings.get('vmmq_adapter_policy_name'):
            module.fail_json(msg=f"vmmq_adapter_policy_name is required when multi_queue_support is true{error_suffix}")


def validate_sriov_settings(module, sriov_settings, name=None):
    """
    Validate SR-IOV connection type settings.
    """
    error_suffix = f" for {name}" if name else ""
    vf_count = sriov_settings.get('vf_count', 64)
    rx_count_per_vf = sriov_settings.get('rx_count_per_vf', 4)
    tx_count_per_vf = sriov_settings.get('tx_count_per_vf', 1)
    comp_count_per_vf = sriov_settings.get('comp_count_per_vf', 5)
    int_count_per_vf = sriov_settings.get('int_count_per_vf', 8)

    if vf_count < 1 or vf_count > 64:
        module.fail_json(msg=f"SR-IOV vf_count must be between 1 and 64{error_suffix}")
    if rx_count_per_vf < 1 or rx_count_per_vf > 8:
        module.fail_json(msg=f"SR-IOV rx_count_per_vf must be between 1 and 8{error_suffix}")
    if tx_count_per_vf < 1 or tx_count_per_vf > 8:
        module.fail_json(msg=f"SR-IOV tx_count_per_vf must be between 1 and 8{error_suffix}")
    if comp_count_per_vf < 1 or comp_count_per_vf > 16:
        module.fail_json(msg=f"SR-IOV comp_count_per_vf must be between 1 and 16{error_suffix}")
    if int_count_per_vf < 1 or int_count_per_vf > 16:
        module.fail_json(msg=f"SR-IOV int_count_per_vf must be between 1 and 16{error_suffix}")


def build_usnic_settings(intersight, policy_cache, module, usnic_settings, organization_name):
    """
    Build USNIC settings for API body.
    """
    usnic_adapter_policy_name = usnic_settings.get('usnic_adapter_policy_name')
    usnic_adapter_policy_moid = get_policy_moid_with_org(
        intersight, policy_cache, module, '/vnic/EthAdapterPolicies',
        usnic_adapter_policy_name, organization_name, 'USNIC Adapter Policy'
    )

    return {
        'Count': usnic_settings.get('count', 0),
        'UsnicAdapterPolicy': usnic_adapter_policy_moid,
        'Cos': usnic_settings.get('cos', 5)
    }


def build_vmq_settings(intersight, policy_cache, module, vmq_settings, organization_name):
    """
    Build VMQ settings for API body.
    """
    multi_queue_support = vmq_settings.get('multi_queue_support', False)

    if multi_queue_support:
        vmmq_adapter_policy_name = vmq_settings.get('vmmq_adapter_policy_name')
        vmmq_adapter_policy_moid = get_policy_moid_with_org(
            intersight, policy_cache, module, '/vnic/EthAdapterPolicies',
            vmmq_adapter_policy_name, organization_name, 'VMMQ Adapter Policy'
        )

        return {
            'Enabled': vmq_settings.get('enabled', True),
            'MultiQueueSupport': True,
            'NumSubVnics': vmq_settings.get('num_sub_vnics', 64),
            'VmmqAdapterPolicy': vmmq_adapter_policy_moid
        }
    else:
        return {
            'Enabled': vmq_settings.get('enabled', True),
            'MultiQueueSupport': False,
            'NumInterrupts': vmq_settings.get('num_interrupts', 16),
            'NumVmqs': vmq_settings.get('num_vmqs', 4)
        }


def build_sriov_settings(sriov_settings):
    """
    Build SR-IOV settings for API body.
    """
    return {
        'VfCount': sriov_settings.get('vf_count', 64),
        'RxCountPerVf': sriov_settings.get('rx_count_per_vf', 4),
        'TxCountPerVf': sriov_settings.get('tx_count_per_vf', 1),
        'CompCountPerVf': sriov_settings.get('comp_count_per_vf', 5),
        'IntCountPerVf': sriov_settings.get('int_count_per_vf', 8),
        'Enabled': sriov_settings.get('enabled', True)
    }


def build_cdn_config(config_source):
    """
    Build CDN configuration for API body.
    """
    cdn_config = {
        'Source': config_source.get('cdn_source', 'vnic')
    }

    # Add CDN value if specified
    if config_source.get('cdn_value'):
        cdn_config['Value'] = config_source['cdn_value']

    return cdn_config


def get_common_policy_mappings():
    """
    Get common policy mappings shared between vNIC template and LAN connectivity policy.
    """
    return {
        'eth_qos_policy_name': ('/vnic/EthQosPolicies', 'EthQosPolicy', 'Ethernet QoS Policy'),
        'eth_adapter_policy_name': ('/vnic/EthAdapterPolicies', 'EthAdapterPolicy', 'Ethernet Adapter Policy')
    }


def get_iscsi_boot_policy_mapping():
    """
    Get iSCSI boot policy mapping (optional policy).
    """
    return {
        'iscsi_boot_policy_name': ('/vnic/IscsiBootPolicies', 'IscsiBootPolicy', 'iSCSI Boot Policy')
    }


def get_fabric_network_policy_mappings():
    """
    Get fabric network policy mappings for FI-attached configurations.
    """
    return {
        'fabric_eth_network_group_policy_name': (
            '/fabric/EthNetworkGroupPolicies', 'FabricEthNetworkGroupPolicy', 'Fabric Ethernet Network Group Policy'
        ),
        'fabric_eth_network_control_policy_name': (
            '/fabric/EthNetworkControlPolicies', 'FabricEthNetworkControlPolicy', 'Fabric Ethernet Network Control Policy'
        )
    }


def get_mac_pool_policy_mapping():
    """
    Get MAC pool policy mapping.
    """
    return {
        'mac_pool_name': ('/macpool/Pools', 'MacPool', 'MAC Pool')
    }


def get_common_settings_argument_spec():
    """
    Get common connection settings argument specifications.
    """
    return {
        'connection_type': dict(type='str', choices=['none', 'usnic', 'vmq', 'sriov'], default='none'),
        'iscsi_boot_policy_name': dict(type='str'),
        'fabric_eth_network_control_policy_name': dict(type='str'),
        'fabric_eth_network_group_policy_name': dict(type='str'),
        'failover_enabled': dict(type='bool', default=False),
        'switch_id': dict(type='str', choices=['A', 'B'], default='A'),
        'mac_pool_name': dict(type='str'),
        'cdn_value': dict(type='str'),
        'cdn_source': dict(type='str', choices=['vnic', 'user'], default='vnic'),
        'usnic_settings': dict(type='dict', options=dict(
            count=dict(type='int', default=0),
            cos=dict(type='int', choices=[0, 1, 2, 3, 4, 5, 6], default=5),
            usnic_adapter_policy_name=dict(type='str')
        )),
        'vmq_settings': dict(type='dict', options=dict(
            enabled=dict(type='bool', default=True),
            multi_queue_support=dict(type='bool', default=False),
            num_interrupts=dict(type='int', default=16),
            num_vmqs=dict(type='int', default=4),
            num_sub_vnics=dict(type='int', default=64),
            vmmq_adapter_policy_name=dict(type='str')
        )),
        'sriov_settings': dict(type='dict', options=dict(
            enabled=dict(type='bool', default=True),
            vf_count=dict(type='int', default=64),
            rx_count_per_vf=dict(type='int', default=4),
            tx_count_per_vf=dict(type='int', default=1),
            comp_count_per_vf=dict(type='int', default=5),
            int_count_per_vf=dict(type='int', default=8)
        ))
    }


def resolve_policy_moids_from_mappings(intersight, policy_cache, module, config_source, policy_mappings, organization_name):
    """
    Resolve policy MOIDs from configuration using provided policy mappings.
    """
    policy_moids = {}

    for param_name, (resource_path, api_field, policy_type) in policy_mappings.items():
        policy_name = config_source.get(param_name)
        if policy_name:
            policy_moid = get_policy_moid_with_org(
                intersight, policy_cache, module, resource_path, policy_name, organization_name, policy_type
            )
            # Special handling for FabricEthNetworkGroupPolicy which needs to be an array for the API
            if api_field == 'FabricEthNetworkGroupPolicy':
                policy_moids[api_field] = [policy_moid]
            else:
                policy_moids[api_field] = policy_moid

    return policy_moids


def build_connection_settings(intersight, policy_cache, module, config_source, organization_name):
    """
    Build connection type specific settings for API body.
    """
    connection_settings = {}
    connection_type = config_source.get('connection_type', 'none')

    if connection_type == 'usnic':
        usnic_settings = config_source.get('usnic_settings', {})
        connection_settings['UsnicSettings'] = build_usnic_settings(intersight, policy_cache, module, usnic_settings, organization_name)
    elif connection_type == 'vmq':
        vmq_settings = config_source.get('vmq_settings', {})
        connection_settings['VmqSettings'] = build_vmq_settings(intersight, policy_cache, module, vmq_settings, organization_name)
    elif connection_type == 'sriov':
        sriov_settings = config_source.get('sriov_settings', {})
        connection_settings['SriovSettings'] = build_sriov_settings(sriov_settings)

    return connection_settings



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# Copyright (c) 2025 Bill Peck, (@p3ck) - Additional changes to make it work for ansible
#

from __future__ import absolute_import, division, print_function
__metaclass__ = type


import re
try:
    from dateutil.parser import parse
    from azure.mgmt.eventgrid.models import (
        EventSubscription,
        EventSubscriptionUpdateParameters,
        WebHookEventSubscriptionDestination,
        RetryPolicy,
        EventHubEventSubscriptionDestination,
        HybridConnectionEventSubscriptionDestination,
        ServiceBusQueueEventSubscriptionDestination,
        ServiceBusTopicEventSubscriptionDestination,
        AzureFunctionEventSubscriptionDestination,
        StorageBlobDeadLetterDestination,
        EventSubscriptionFilter,
        EventSubscriptionIdentity,
        DeliveryWithResourceIdentity,
        DeadLetterWithResourceIdentity,
        StorageQueueEventSubscriptionDestination)
except ImportError:
    # This is handled in azure_rm_common
    pass


IDENTITY_NO_IDENTITY = "NoIdentity"
IDENTITY_NONE = "None"
IDENTITY_SYSTEM_ASSIGNED = "SystemAssigned"
IDENTITY_USER_ASSIGNED = "UserAssigned"
IDENTITY_MIXED_MODE = "SystemAssigned, UserAssigned"

WEBHOOK_DESTINATION = "webhook"
EVENTHUB_DESTINATION = "eventhub"
STORAGEQUEUE_DESTINATION = "storagequeue"
HYBRIDCONNECTION_DESTINATION = "hybridconnection"
SERVICEBUSQUEUE_DESTINATION = "servicebusqueue"
SERVICEBUSTOPIC_DESTINATION = "servicebustopic"
AZUREFUNCTION_DESTINATION = "azurefunction"


class ConfigError(Exception):
    def __init__(self, message, **kwargs):
        self.__dict__.update(kwargs)
        super(ConfigError, self).__init__(message)


eventgrid_subscription_module_arg_spec = dict(
    azure_active_directory_application_id_or_uri=dict(
        type='str',
        required=False,
    ),
    azure_active_directory_tenant_id=dict(
        type='str',
        required=False,
    ),
    deadletter_endpoint=dict(
        type='str',
        required=False,
    ),
    delivery_attribute_mapping=dict(
        type='list',
        elements='str',
        required=False,
    ),
    endpoint=dict(
        type='str',
        required=False,
    ),
    endpoint_type=dict(
        type='str',
        choices=['azurefunction',
                 'eventhub',
                 'hybridconnection',
                 'servicebusqueue',
                 'servicebustopic',
                 'storagequeue',
                 'webhook'],
        default='webhook',
    ),
    event_delivery_schema=dict(
        type='str',
        choices=['CloudEventSchemaV1_0', 'CustomInputSchema', 'EventGridSchema'],
        required=False,
    ),
    event_ttl=dict(
        type='int',
        default=1440,
    ),
    expiration_date=dict(
        type='str',
        required=False,
    ),
    labels=dict(
        type='list',
        elements='str',
        required=False,
    ),
    max_delivery_attempts=dict(
        type='int',
        default=30,
        required=False,
    ),
    max_events_per_batch=dict(
        type='int',
        required=False,
    ),
    preferred_batch_size_in_kilobytes=dict(
        type='int',
        required=False,
    ),
    storage_queue_msg_ttl=dict(
        type='int',
        required=False,
    ),
    advanced_filter=dict(
        type='list',
        elements='str',
        required=False,
    ),
    enable_advanced_filtering_on_arrays=dict(
        type='bool',
        required=False,
    ),
    included_event_types=dict(
        type='list',
        elements='str',
        required=False,
    ),
    subject_begins_with=dict(
        type='str',
        required=False,
    ),
    subject_case_sensitive=dict(
        type='bool',
        required=False,
    ),
    subject_ends_with=dict(
        type='str',
        required=False,
    ),
)


def _validate_retry_policy(max_delivery_attempts, event_ttl):
    if max_delivery_attempts < 1 or max_delivery_attempts > 30:
        raise ConfigError('max_delivery_attempts should be a number between 1 and 30.')

    if event_ttl < 1 or event_ttl > 1440:
        raise ConfigError('event_ttl should be a number between 1 and 1440.')


def _validate_destination_attribute(endpoint_type, storage_queue_msg_ttl=None, delivery_attribute_mapping=None):
    isStorageQueueDestination = endpoint_type is not None and endpoint_type.lower() == STORAGEQUEUE_DESTINATION.lower()

    if not isStorageQueueDestination and storage_queue_msg_ttl is not None:
        raise ConfigError('usage error: storage_queue_msg_ttl is only applicable for endpoint type StorageQueue.')

    if isStorageQueueDestination and delivery_attribute_mapping is not None:
        raise ConfigError('usage error: delivery_attribute_mapping is not applicable for endpoint type StorageQueue.')


def _get_storage_queue_destination(endpoint, storage_queue_msg_ttl):
    # Supplied endpoint would be in the following format:
    # /subscriptions/.../storageAccounts/sa1/queueServices/default/queues/{queueName}))
    # and we need to break it up into:
    # /subscriptions/.../storageAccounts/sa1 and queueName
    queue_items = re.split(
        "/queueServices/default/queues/", endpoint, flags=re.IGNORECASE)

    if len(queue_items) != 2 or queue_items[0] is None or queue_items[1] is None:
        raise ConfigError('Argument Error: Expected format of endpoint for storage queue is:' +
                          '/subscriptions/id/resourceGroups/rg/providers/Microsoft.Storage/' +
                          'storageAccounts/sa1/queueServices/default/queues/queueName')

    if storage_queue_msg_ttl is not None:
        destination = StorageQueueEventSubscriptionDestination(
            resource_id=queue_items[0],
            queue_name=queue_items[1],
            queue_message_time_to_live_in_seconds=storage_queue_msg_ttl)
    else:
        destination = StorageQueueEventSubscriptionDestination(
            resource_id=queue_items[0],
            queue_name=queue_items[1])
    return destination


def _get_endpoint_destination(
        endpoint_type,
        endpoint,
        max_events_per_batch,
        preferred_batch_size_in_kilobytes,
        azure_active_directory_tenant_id,
        azure_active_directory_application_id_or_uri,
        storage_queue_msg_ttl,
        delivery_attribute_mapping):

    if endpoint_type.lower() == WEBHOOK_DESTINATION.lower():
        destination = WebHookEventSubscriptionDestination(
            endpoint_url=endpoint,
            max_events_per_batch=max_events_per_batch,
            preferred_batch_size_in_kilobytes=preferred_batch_size_in_kilobytes,
            azure_active_directory_tenant_id=azure_active_directory_tenant_id,
            azure_active_directory_application_id_or_uri=azure_active_directory_application_id_or_uri,
            delivery_attribute_mappings=delivery_attribute_mapping)
    elif endpoint_type.lower() == EVENTHUB_DESTINATION.lower():
        destination = EventHubEventSubscriptionDestination(
            resource_id=endpoint,
            delivery_attribute_mappings=delivery_attribute_mapping)
    elif endpoint_type.lower() == HYBRIDCONNECTION_DESTINATION.lower():
        destination = HybridConnectionEventSubscriptionDestination(
            resource_id=endpoint,
            delivery_attribute_mappings=delivery_attribute_mapping)
    elif endpoint_type.lower() == STORAGEQUEUE_DESTINATION.lower():
        destination = _get_storage_queue_destination(endpoint, storage_queue_msg_ttl)
    elif endpoint_type.lower() == SERVICEBUSQUEUE_DESTINATION.lower():
        destination = ServiceBusQueueEventSubscriptionDestination(
            resource_id=endpoint,
            delivery_attribute_mappings=delivery_attribute_mapping)
    elif endpoint_type.lower() == SERVICEBUSTOPIC_DESTINATION.lower():
        destination = ServiceBusTopicEventSubscriptionDestination(
            resource_id=endpoint,
            delivery_attribute_mappings=delivery_attribute_mapping)
    elif endpoint_type.lower() == AZUREFUNCTION_DESTINATION.lower():
        destination = AzureFunctionEventSubscriptionDestination(
            resource_id=endpoint,
            max_events_per_batch=max_events_per_batch,
            preferred_batch_size_in_kilobytes=preferred_batch_size_in_kilobytes,
            delivery_attribute_mappings=delivery_attribute_mapping)
    return destination


def _get_event_subscription_identity_type(identity_type_name):
    result = None
    if identity_type_name.lower() == IDENTITY_SYSTEM_ASSIGNED.lower():
        result = IDENTITY_SYSTEM_ASSIGNED

    return result


def _get_deadletter_destination(deadletter_endpoint):
    blob_items = re.split(
        "/blobServices/default/containers/", deadletter_endpoint, flags=re.IGNORECASE)

    if len(blob_items) != 2 or blob_items[0] is None or blob_items[1] is None:
        raise ConfigError('Argument Error: Expected format of deadletter_endpoint is:' +
                          '/subscriptions/id/resourceGroups/rg/providers/Microsoft.Storage/' +
                          'storageAccounts/sa1/blobServices/default/containers/containerName')

    return StorageBlobDeadLetterDestination(resource_id=blob_items[0], blob_container_name=blob_items[1])


def _get_tenant_id(destination, destination_with_resource_identity):
    tenant_id = None

    if destination is not None and hasattr(destination, 'azure_active_directory_tenant_id'):
        tenant_id = destination.azure_active_directory_tenant_id
    elif destination_with_resource_identity is not None and hasattr(destination_with_resource_identity, 'azure_active_directory_tenant_id'):  # nopep8: E501
        tenant_id = destination_with_resource_identity.azure_active_directory_tenant_id

    return tenant_id


def _get_application_id(destination, destination_with_resource_identity):
    application_id = None

    if destination is not None and hasattr(destination, 'azure_active_directory_application_id_or_uri'):
        application_id = destination.azure_active_directory_application_id_or_uri
    elif destination_with_resource_identity is not None and hasattr(destination_with_resource_identity, 'azure_active_directory_application_id_or_uri'):  # nopep8: E501
        application_id = destination_with_resource_identity.azure_active_directory_application_id_or_uri
    return application_id


def _set_event_subscription_destination(
        destination,
        storage_queue_msg_ttl=None,
        delivery_attribute_mapping=None):

    endpoint_type = destination.endpoint_type
    if endpoint_type.lower() == WEBHOOK_DESTINATION.lower():
        if delivery_attribute_mapping is not None:
            destination.delivery_attribute_mappings = delivery_attribute_mapping
    elif endpoint_type.lower() == EVENTHUB_DESTINATION.lower():
        if delivery_attribute_mapping is not None:
            destination.delivery_attribute_mappings = delivery_attribute_mapping
    elif endpoint_type.lower() == HYBRIDCONNECTION_DESTINATION.lower():
        if delivery_attribute_mapping is not None:
            destination.delivery_attribute_mappings = delivery_attribute_mapping
    elif endpoint_type.lower() == STORAGEQUEUE_DESTINATION.lower():
        if storage_queue_msg_ttl is not None:
            destination.queue_message_time_to_live_in_seconds = storage_queue_msg_ttl
    elif endpoint_type.lower() == SERVICEBUSQUEUE_DESTINATION.lower():
        if delivery_attribute_mapping is not None:
            destination.delivery_attribute_mappings = delivery_attribute_mapping
    elif endpoint_type.lower() == SERVICEBUSTOPIC_DESTINATION.lower():
        if delivery_attribute_mapping is not None:
            destination.delivery_attribute_mappings = delivery_attribute_mapping
    elif endpoint_type.lower() == AZUREFUNCTION_DESTINATION.lower():
        if delivery_attribute_mapping is not None:
            destination.delivery_attribute_mappings = delivery_attribute_mapping
    return destination


def _validate_and_update_destination(endpoint_type, destination, storage_queue_msg_ttl, delivery_attribute_mapping):
    _validate_destination_attribute(
        endpoint_type,
        storage_queue_msg_ttl,
        delivery_attribute_mapping)

    _set_event_subscription_destination(
        destination,
        storage_queue_msg_ttl,
        delivery_attribute_mapping)


def _update_event_subscription_filter(
        event_subscription_filter,
        subject_begins_with=None,
        subject_ends_with=None,
        included_event_types=None,
        enable_advanced_filtering_on_arrays=None,
        advanced_filter=None):

    if subject_begins_with is not None:
        event_subscription_filter.subject_begins_with = subject_begins_with

    if subject_ends_with is not None:
        event_subscription_filter.subject_ends_with = subject_ends_with

    if included_event_types is not None:
        event_subscription_filter.included_event_types = included_event_types

    if enable_advanced_filtering_on_arrays is not None:
        event_subscription_filter.enable_advanced_filtering_on_arrays = enable_advanced_filtering_on_arrays

    if advanced_filter is not None:
        event_subscription_filter.advanced_filters = advanced_filter


def get_event_subscription_info(endpoint=None,
                                endpoint_type=WEBHOOK_DESTINATION,
                                included_event_types=None,
                                subject_begins_with=None,
                                subject_ends_with=None,
                                is_subject_case_sensitive=False,
                                max_delivery_attempts=30,
                                event_ttl=1440,
                                max_events_per_batch=None,
                                preferred_batch_size_in_kilobytes=None,
                                event_delivery_schema=None,
                                deadletter_endpoint=None,
                                labels=None,
                                expiration_date=None,
                                advanced_filter=None,
                                azure_active_directory_tenant_id=None,
                                azure_active_directory_application_id_or_uri=None,
                                delivery_identity=None,
                                delivery_identity_endpoint=None,
                                delivery_identity_endpoint_type=None,
                                deadletter_identity=None,
                                deadletter_identity_endpoint=None,
                                storage_queue_msg_ttl=None,
                                enable_advanced_filtering_on_arrays=None,
                                delivery_attribute_mapping=None):

    normalized_endpoint_type = endpoint_type.lower()
    normalized_webhook_destination = WEBHOOK_DESTINATION.lower()
    normalized_azure_function_destination = AZUREFUNCTION_DESTINATION.lower()

    if deadletter_endpoint is not None and deadletter_identity_endpoint is not None:
        raise ConfigError('usage error: either deadletter_endpoint or deadletter_identity_endpoint '
                          'should be specified at one time, not both')

    if included_event_types is not None and len(included_event_types) == 1 and included_event_types[0].lower() == 'all':
        raise ConfigError('The usage of \"All\" for included_event_types is not allowed starting from Azure Event Grid'
                          ' API Version 2019-02-01-preview.')

    # Construct RetryPolicy based on max_delivery_attempts and event_ttl
    _validate_retry_policy(max_delivery_attempts, event_ttl)
    retry_policy = RetryPolicy(max_delivery_attempts=max_delivery_attempts, event_time_to_live_in_minutes=event_ttl)

    if max_events_per_batch is not None:
        if normalized_endpoint_type not in (normalized_webhook_destination, normalized_azure_function_destination):
            raise ConfigError('usage error: max_events_per_batch is applicable only for '
                              'endpoint types WebHook and AzureFunction.')
        if max_events_per_batch > 5000:
            raise ConfigError('usage error: max_events_per_batch must be a number between 1 and 5000.')

    if preferred_batch_size_in_kilobytes is not None:
        if normalized_endpoint_type not in (normalized_webhook_destination, normalized_azure_function_destination):
            raise ConfigError('usage error: preferred_batch_size_in_kilobytes is applicable only for '
                              'endpoint types WebHook and AzureFunction.')
        if preferred_batch_size_in_kilobytes > 1024:
            raise ConfigError('usage error: preferred_batch_size_in_kilobytes must be a number '
                              'between 1 and 1024.')

    if azure_active_directory_tenant_id is not None:
        if normalized_endpoint_type != normalized_webhook_destination:
            raise ConfigError('usage error: azure_active_directory_tenant_id is applicable only for '
                              'endpoint types WebHook.')
        if azure_active_directory_application_id_or_uri is None:
            raise ConfigError('usage error: azure_active_directory_application_id_or_uri is missing. '
                              'It should include an Azure Active Directory Application Id or Uri.')

    if azure_active_directory_application_id_or_uri is not None:
        if normalized_endpoint_type != normalized_webhook_destination:
            raise ConfigError('usage error: azure_active_directory_application_id_or_uri is applicable only for '
                              'endpoint types WebHook.')
        if azure_active_directory_tenant_id is None:
            raise ConfigError('usage error: azure_active_directory_tenant_id is missing. '
                              'It should include an Azure Active Directory Tenant Id.')

    condition1 = deadletter_identity is not None and deadletter_identity_endpoint is None
    condition2 = deadletter_identity is None and deadletter_identity_endpoint is not None
    if condition1 or condition2:
        raise ConfigError('usage error: one or more deadletter identity information is missing. If '
                          'deadletter_identity is specified, deadletter_identity_endpoint should be specified.')

    tennant_id = None
    application_id = None

    condition1 = endpoint_type is not None and normalized_endpoint_type == normalized_webhook_destination
    condition2 = delivery_identity_endpoint_type is not None and \
        delivery_identity_endpoint_type.lower() == normalized_webhook_destination
    if condition1 or condition2:
        tennant_id = azure_active_directory_tenant_id
        application_id = azure_active_directory_application_id_or_uri

    destination = None
    delivery_with_resource_identity = None
    if endpoint is not None:
        _validate_destination_attribute(endpoint_type, storage_queue_msg_ttl, delivery_attribute_mapping)
        destination = _get_endpoint_destination(
            endpoint_type,
            endpoint,
            max_events_per_batch,
            preferred_batch_size_in_kilobytes,
            tennant_id,
            application_id,
            storage_queue_msg_ttl,
            delivery_attribute_mapping)
    elif delivery_identity_endpoint is not None:
        identity_type_name = _get_event_subscription_identity_type(delivery_identity)
        delivery_identity_info = EventSubscriptionIdentity(type=identity_type_name)
        _validate_destination_attribute(
            delivery_identity_endpoint_type,
            storage_queue_msg_ttl,
            delivery_attribute_mapping)
        destination_with_identity = _get_endpoint_destination(
            delivery_identity_endpoint_type,
            delivery_identity_endpoint,
            max_events_per_batch,
            preferred_batch_size_in_kilobytes,
            tennant_id,
            application_id,
            storage_queue_msg_ttl,
            delivery_attribute_mapping)
        delivery_with_resource_identity = DeliveryWithResourceIdentity(
            identity=delivery_identity_info,
            destination=destination_with_identity)

    event_subscription_filter = EventSubscriptionFilter(
        subject_begins_with=subject_begins_with,
        subject_ends_with=subject_ends_with,
        included_event_types=included_event_types,
        is_subject_case_sensitive=is_subject_case_sensitive,
        enable_advanced_filtering_on_arrays=enable_advanced_filtering_on_arrays,
        advanced_filters=advanced_filter)

    deadletter_destination = None
    if deadletter_endpoint is not None:
        deadletter_destination = _get_deadletter_destination(deadletter_endpoint)

    deadletter_with_resource_identity = None

    if deadletter_identity_endpoint is not None:
        deadletter_destination_with_identity = _get_deadletter_destination(deadletter_identity_endpoint)
        deadletter_identity_type_name = _get_event_subscription_identity_type(deadletter_identity)
        deadletter_delivery_identity_info = EventSubscriptionIdentity(type=deadletter_identity_type_name)
        deadletter_with_resource_identity = DeadLetterWithResourceIdentity(
            identity=deadletter_delivery_identity_info,
            dead_letter_destination=deadletter_destination_with_identity)

    if expiration_date is not None:
        expiration_date = parse(expiration_date)

    event_subscription_info = EventSubscription(
        destination=destination,
        filter=event_subscription_filter,
        labels=labels,
        event_delivery_schema=event_delivery_schema,
        retry_policy=retry_policy,
        expiration_time_utc=expiration_date,
        dead_letter_destination=deadletter_destination,
        delivery_with_resource_identity=delivery_with_resource_identity,
        dead_letter_with_resource_identity=deadletter_with_resource_identity)

    return event_subscription_info


def update_event_subscription_internal(instance,
                                       endpoint=None,
                                       endpoint_type=WEBHOOK_DESTINATION,
                                       subject_begins_with=None,
                                       subject_ends_with=None,
                                       included_event_types=None,
                                       advanced_filter=None,
                                       labels=None,
                                       deadletter_endpoint=None,
                                       delivery_identity=None,
                                       delivery_identity_endpoint=None,
                                       delivery_identity_endpoint_type=None,
                                       deadletter_identity=None,
                                       deadletter_identity_endpoint=None,
                                       storage_queue_msg_ttl=None,
                                       enable_advanced_filtering_on_arrays=None,
                                       delivery_attribute_mapping=None):

    if (endpoint_type is not None and
            endpoint_type.lower() != WEBHOOK_DESTINATION.lower() and
            endpoint is None):
        raise ConfigError('Invalid usage: Since endpoint_type is specified, a valid endpoint must also be specified.')

    current_destination = instance.destination
    current_filter = instance.filter
    current_event_delivery_schema = instance.event_delivery_schema
    current_retry_policy = instance.retry_policy
    current_destination_with_resource_identity = None
    current_destination2 = None

    if instance.delivery_with_resource_identity is not None:
        current_destination2 = instance.delivery_with_resource_identity
        current_destination_with_resource_identity = instance.delivery_with_resource_identity.destination

    tenant_id = _get_tenant_id(current_destination, current_destination_with_resource_identity)
    application_id = _get_application_id(current_destination, current_destination_with_resource_identity)

    # for the update path, endpoint_type can be None but it does not mean that this is webhook,
    # as it can be other types too.
    current_max_events_per_batch = 0
    current_preferred_batch_size_in_kilobytes = 0

    if current_destination is not None and (current_destination.endpoint_type.lower() == WEBHOOK_DESTINATION.lower() or current_destination.endpoint_type.lower() == AZUREFUNCTION_DESTINATION.lower()):  # nopep8: E501
        current_max_events_per_batch = current_destination.max_events_per_batch
        current_preferred_batch_size_in_kilobytes = current_destination.preferred_batch_size_in_kilobytes
    elif current_destination_with_resource_identity is not None and (current_destination_with_resource_identity.endpoint_type.lower() == WEBHOOK_DESTINATION.lower() or current_destination_with_resource_identity.endpoint_type.lower() == AZUREFUNCTION_DESTINATION.lower()):  # nopep8: E501
        current_max_events_per_batch = current_destination_with_resource_identity.max_events_per_batch
        current_preferred_batch_size_in_kilobytes = current_destination_with_resource_identity.preferred_batch_size_in_kilobytes  # nopep8: E501

    updated_destination = None
    updated_delivery_with_resource_identity = None

    # if endpoint and delivery_identity_endpoint is not specified then use the instance value
    if endpoint is None and delivery_identity_endpoint is None:
        if current_destination is not None:
            _validate_and_update_destination(
                current_destination.endpoint_type,
                current_destination,
                storage_queue_msg_ttl,
                delivery_attribute_mapping)
            updated_destination = current_destination
        elif current_destination_with_resource_identity is not None:
            _validate_and_update_destination(
                current_destination_with_resource_identity.endpoint_type,
                current_destination_with_resource_identity,
                storage_queue_msg_ttl,
                delivery_attribute_mapping)
            updated_delivery_with_resource_identity = current_destination2
    elif endpoint is not None:
        # If this is the update path, the user does not
        # have to specify the endpoint type so it might
        # be None
        if endpoint_type is None:
            endpoint_type = current_destination.endpoint_type

        _validate_destination_attribute(
            endpoint_type,
            storage_queue_msg_ttl,
            delivery_attribute_mapping)
        updated_destination = _get_endpoint_destination(
            endpoint_type,
            endpoint,
            current_max_events_per_batch,
            current_preferred_batch_size_in_kilobytes,
            tenant_id,
            application_id,
            storage_queue_msg_ttl,
            delivery_attribute_mapping)
    elif delivery_identity_endpoint is not None:
        destination_with_identity = _get_endpoint_destination(
            delivery_identity_endpoint_type,
            delivery_identity_endpoint,
            0,
            0,
            tenant_id,
            application_id,
            storage_queue_msg_ttl,
            delivery_attribute_mapping)

        identity_type_name = _get_event_subscription_identity_type(delivery_identity)
        delivery_identity_info = EventSubscriptionIdentity(type=identity_type_name)

        updated_delivery_with_resource_identity = DeliveryWithResourceIdentity(
            identity=delivery_identity_info,
            destination=destination_with_identity)

    updated_deadletter_destination = None
    if deadletter_endpoint is not None:
        updated_deadletter_destination = _get_deadletter_destination(deadletter_endpoint)

    updated_deadletter_with_resource_identity = None
    if deadletter_identity_endpoint is not None:
        deadletter_destination_with_identity = _get_deadletter_destination(deadletter_identity_endpoint)
        deadletter_identity_type_name = _get_event_subscription_identity_type(deadletter_identity)
        deadletter_delivery_identity_info = EventSubscriptionIdentity(type=deadletter_identity_type_name)
        updated_deadletter_with_resource_identity = DeadLetterWithResourceIdentity(
            identity=deadletter_delivery_identity_info,
            dead_letter_destination=deadletter_destination_with_identity)

    _update_event_subscription_filter(
        current_filter,
        subject_begins_with,
        subject_ends_with,
        included_event_types,
        enable_advanced_filtering_on_arrays,
        advanced_filter)
    updated_filter = current_filter

    updated_labels = None
    if labels is not None:
        updated_labels = labels

    params = EventSubscriptionUpdateParameters(
        destination=updated_destination,
        filter=updated_filter,
        labels=updated_labels,
        retry_policy=current_retry_policy,
        dead_letter_destination=updated_deadletter_destination,
        event_delivery_schema=current_event_delivery_schema,
        delivery_with_resource_identity=updated_delivery_with_resource_identity,
        dead_letter_with_resource_identity=updated_deadletter_with_resource_identity)

    return params



# -*- coding: utf-8 -*-
# Copyright 2024 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The nxos vrf_address_family fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import utils

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.argspec.vrf_address_family.vrf_address_family import (
    Vrf_address_familyArgs,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.rm_templates.vrf_address_family import (
    Vrf_address_familyTemplate,
)


class Vrf_address_familyFacts(object):
    """The nxos vrf_address_family facts class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = Vrf_address_familyArgs.argument_spec

    def get_config(self, connection):
        """Get the configuration from the device"""

        return connection.get("show running-config | section ^vrf")

    def _flatten_config(self, data):
        """Flatten contexts in the vrf address family
            running-config for easier parsing.
        :param data: string running-config
        :returns: flattened running config
        """

        dataLines = data.split("\n")
        curData = ""

        for line in dataLines:
            if "vrf context" in line:
                curData = line
            elif "address-family" in line:
                dataLines[dataLines.index(line)] = curData + " " + line

        return "\n".join(dataLines)

    def _parse_vrf_af(self, data: dict):
        """Parse the vrf address family data
        :param data: dict of vrf address family data
        :returns: argspec compliant list
        """
        if not data:
            return []

        vrf_lists = list(data.values())
        for item in vrf_lists:
            if "address_families" in item:
                item["address_families"] = list(item["address_families"].values())
        return vrf_lists

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for Vrf_address_family network resource

        :param connection: the device connection
        :param ansible_facts: Facts dictionary
        :param data: previously collected conf

        :rtype: dictionary
        :returns: facts
        """
        facts = {}
        vrfObjs = []

        if not data:
            data = self.get_config(connection)

        flattened_data = self._flatten_config(data)

        # parse native config using the Vrf_address_family template
        vrf_address_family_parser = Vrf_address_familyTemplate(
            lines=flattened_data.splitlines(),
            module=self._module,
        )
        parsed_data = vrf_address_family_parser.parse()
        vrfObjs = self._parse_vrf_af(parsed_data)

        ansible_facts["ansible_network_resources"].pop("vrf_address_family", None)

        params = utils.remove_empties(
            vrf_address_family_parser.validate_config(
                self.argument_spec,
                {"config": vrfObjs},
                redact=True,
            ),
        )
        facts["vrf_address_family"] = params.get("config", [])
        ansible_facts["ansible_network_resources"].update(facts)

        return ansible_facts



#!/usr/bin/env python3
"""
Ejecuta todos los playbooks de Ansible y genera un reporte de exitosos y fallidos.

Uso:
    python3 run_playbooks.py
    python3 run_playbooks.py --credentials playbooks/credentials.yml
    python3 run_playbooks.py --dir playbooks --timeout 120
"""

import argparse
import os
import re
import subprocess
import sys
from datetime import datetime
from pathlib import Path

LOGS_DIR = Path("playbooks_logs")
SKIP_FILES = {"credentials.template", "credentials.yml", "hosts"}
SKIP_EXTENSIONS = {".template"}

# Archivos de vars que deben existir para que el playbook corra
# Si no existen se saltea con razon clara
REQUIRED_VAR_FILES_PATTERN = re.compile(
    r"file:\s+['\"]?([^\s'\"#]+\.ya?ml)['\"]?|"
    r"vars_files:\s*\n(?:\s+-\s+['\"]?([^\s'\"#]+\.ya?ml)['\"]?\n)+",
    re.MULTILINE,
)


def find_playbooks(directory: Path) -> list[Path]:
    playbooks = []
    for path in sorted(directory.iterdir()):
        if path.is_dir():
            continue
        if path.name in SKIP_FILES:
            continue
        if path.suffix in SKIP_EXTENSIONS:
            continue
        if path.suffix not in (".yml", ".yaml"):
            continue
        # Solo archivos que tengan 'hosts:' (son playbooks, no vars files)
        try:
            content = path.read_text(encoding="utf-8", errors="ignore")
            if "hosts:" not in content:
                continue
        except Exception:
            continue
        playbooks.append(path)
    return playbooks


def missing_var_files(playbook: Path) -> list[str]:
    """Devuelve lista de var files referenciados pero no encontrados en disco."""
    try:
        content = playbook.read_text(encoding="utf-8", errors="ignore")
    except Exception:
        return []

    missing = []
    for match in REQUIRED_VAR_FILES_PATTERN.finditer(content):
        fname = match.group(1) or match.group(2)
        if not fname:
            continue
        # Buscar relativo al directorio del playbook
        candidate = playbook.parent / fname
        if not candidate.exists():
            missing.append(fname)
    return missing


def run_playbook(
    playbook: Path,
    extra_vars_file: str | None,
    timeout: int,
) -> tuple[str, int, str]:
    """
    Corre un playbook. Devuelve (status, exit_code, output).
    status: 'PASS' | 'FAIL' | 'TIMEOUT' | 'SKIP'
    """
    cmd = ["ansible-playbook", "-i", "localhost,", str(playbook)]
    if extra_vars_file:
        cmd += ["-e", f"@{extra_vars_file}"]

    try:
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            timeout=timeout,
            cwd=playbook.parent.parent,  # raiz del repo
        )
        output = result.stdout + result.stderr
        status = "PASS" if result.returncode == 0 else "FAIL"
        return status, result.returncode, output

    except subprocess.TimeoutExpired:
        return "TIMEOUT", -1, f"Timeout ({timeout}s)"

    except FileNotFoundError:
        return "FAIL", -1, "ansible-playbook no encontrado. Instala Ansible."

    except Exception as e:
        return "FAIL", -1, str(e)


def extract_error(output: str) -> str:
    """Extrae la primera linea de error relevante del output."""
    for line in output.splitlines():
        if any(k in line for k in ("fatal:", "ERROR!", "error:", "Exception")):
            return line.strip()[:200]
    return output.strip().splitlines()[-1][:200] if output.strip() else "sin detalle"


def save_log(playbook_name: str, output: str, exit_code: int) -> Path:
    LOGS_DIR.mkdir(exist_ok=True)
    ts = datetime.now().strftime("%Y%m%d_%H%M%S")
    log_path = LOGS_DIR / f"{playbook_name}_{ts}.log"
    log_path.write_text(
        f"Playbook : {playbook_name}\n"
        f"Fecha    : {datetime.now()}\n"
        f"Exit code: {exit_code}\n"
        f"{'='*60}\n"
        f"{output}\n",
        encoding="utf-8",
    )
    return log_path


def main():
    parser = argparse.ArgumentParser(description="Corre todos los playbooks y genera reporte")
    parser.add_argument("--dir", default="playbooks", help="Directorio de playbooks")
    parser.add_argument("--credentials", default=None, help="Archivo de credenciales (extra vars)")
    parser.add_argument("--timeout", type=int, default=120, help="Timeout por playbook en segundos")
    parser.add_argument("--report", default="playbooks_report.txt", help="Archivo de reporte")
    args = parser.parse_args()

    pb_dir = Path(args.dir)
    if not pb_dir.exists():
        print(f"ERROR: directorio '{pb_dir}' no existe.")
        sys.exit(1)

    creds = args.credentials
    if creds and not Path(creds).exists():
        print(f"ADVERTENCIA: credenciales '{creds}' no encontradas, se omiten.")
        creds = None

    playbooks = find_playbooks(pb_dir)
    total = len(playbooks)
    if total == 0:
        print("No se encontraron playbooks.")
        sys.exit(0)

    print(f"\n{'='*60}")
    print(f"  Meraki Playbook Runner")
    print(f"  Playbooks : {total}")
    print(f"  Credenciales: {creds or '(ninguna)'}")
    print(f"  Timeout   : {args.timeout}s")
    print(f"  Reporte   : {args.report}")
    print(f"{'='*60}\n")

    passed, failed, skipped = [], [], []

    for i, pb in enumerate(playbooks, 1):
        name = pb.name
        prefix = f"[{i:>3}/{total}]"

        # Verificar dependencias antes de correr
        missing = missing_var_files(pb)
        if missing:
            reason = f"Vars file(s) no encontrado(s): {', '.join(missing)}"
            print(f"{prefix} SKIP  {name}")
            print(f"          {reason}")
            skipped.append({"name": name, "reason": reason})
            continue

        print(f"{prefix} RUN   {name} ...", end=" ", flush=True)
        status, exit_code, output = run_playbook(pb, creds, args.timeout)

        if status == "PASS":
            print("PASS")
            passed.append({"name": name})
        else:
            error = extract_error(output)
            log_path = save_log(name, output, exit_code)
            print(f"{status}")
            print(f"          {error}")
            print(f"          Log: {log_path}")
            failed.append({"name": name, "status": status, "error": error, "log": str(log_path)})

    # Generar reporte
    lines = [
        "=" * 60,
        "  REPORTE DE EJECUCION DE PLAYBOOKS",
        f"  Fecha   : {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}",
        "=" * 60,
        f"  Total   : {total}",
        f"  Exitosos: {len(passed)}",
        f"  Fallidos: {len(failed)}",
        f"  Saltados: {len(skipped)} (dependencias faltantes)",
        "=" * 60,
        "",
    ]

    if passed:
        lines += ["EXITOSOS", "-" * 60]
        for r in passed:
            lines.append(f"  OK   {r['name']}")
        lines.append("")

    if failed:
        lines += ["FALLIDOS", "-" * 60]
        for r in failed:
            lines.append(f"  FAIL [{r['status']}] {r['name']}")
            lines.append(f"       Error: {r['error']}")
            lines.append(f"       Log  : {r['log']}")
        lines.append("")

    if skipped:
        lines += ["SALTADOS (dependencia faltante)", "-" * 60]
        for r in skipped:
            lines.append(f"  SKIP {r['name']}")
            lines.append(f"       {r['reason']}")
        lines.append("")

    report = "\n".join(lines)
    print("\n" + report)
    Path(args.report).write_text(report, encoding="utf-8")
    print(f"Reporte guardado en: {args.report}")

    sys.exit(0 if not failed else 1)


if __name__ == "__main__":
    main()



# Copyright (c) 2023, Poh Wei Sheng <weisheng-p@hotmail.sg>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
name: github_app_access_token
author:
  - Poh Wei Sheng (@weisheng-p)
  - Bruno Lavoie (@blavoie)
short_description: Obtain short-lived Github App Access tokens
version_added: '8.2.0'
requirements:
  - jwt (https://github.com/GehirnInc/python-jwt) OR
  - PyJWT (https://pypi.org/project/PyJWT/) AND cryptography (https://pypi.org/project/cryptography/)
description:
  - This generates a Github access token that can be used with a C(git) command, if you use a Github App.
options:
  key_path:
    description:
      - Path to your private key.
      - Either O(key_path) or O(private_key) must be specified.
    type: path
  app_id:
    description:
      - Your GitHub App ID, you can find this in the Settings page.
    required: true
    type: str
  installation_id:
    description:
      - The installation ID that contains the git repository you would like access to.
      - As of 2023-12-24, this can be found at Settings page > Integrations > Application. The last part of the URL in the
        configure button is the installation ID.
      - Alternatively, you can use PyGithub (U(https://github.com/PyGithub/PyGithub)) to get your installation ID.
    required: true
    type: str
  private_key:
    description:
      - GitHub App private key in PEM file format as string.
      - Either O(key_path) or O(private_key) must be specified.
    type: str
    version_added: 10.0.0
  token_expiry:
    description:
      - How long the token should last for in seconds.
    default: 600
    type: int
  github_url:
    description:
      - Base URL for the GitHub API (for GitHub Enterprise Server).
      - "Example: C(https://github-enterprise-server.example.com/api/v3)"
    default: https://api.github.com
    type: str
    version_added: 11.4.0
"""

EXAMPLES = r"""
- name: Get access token to be used for git checkout with app_id=123456, installation_id=64209
  ansible.builtin.git:
    repo: >-
      https://x-access-token:{{ github_token }}@github.com/hidden_user/super-secret-repo.git
    dest: /srv/checkout
  vars:
    github_token: >-
      {{ lookup('community.general.github_app_access_token', key_path='/home/to_your/key',
                app_id='123456', installation_id='64209') }}
"""

RETURN = r"""
_raw:
  description: A one-element list containing your GitHub access token.
  type: list
  elements: str
"""

try:
    import jwt

    HAS_JWT = True
except ImportError:
    HAS_JWT = False

HAS_PYTHON_JWT = False  # vs pyjwt
if HAS_JWT and hasattr(jwt, "JWT"):
    HAS_PYTHON_JWT = True
    from jwt import JWT, jwk_from_pem  # type: ignore[attr-defined]

    jwt_instance = JWT()

try:
    from cryptography.hazmat.primitives import serialization

    HAS_CRYPTOGRAPHY = True
except ImportError:
    HAS_CRYPTOGRAPHY = False


import json
import time
from http import HTTPStatus
from urllib.error import HTTPError

from ansible.errors import AnsibleError, AnsibleOptionsError
from ansible.module_utils.urls import open_url
from ansible.plugins.lookup import LookupBase
from ansible.utils.display import Display

from ansible_collections.community.general.plugins.plugin_utils._lookup import check_for_no_terms

display = Display()


class PythonJWT:
    @staticmethod
    def read_key(path, private_key=None):
        try:
            if private_key:
                return jwk_from_pem(private_key.encode("utf-8"))
            with open(path, "rb") as pem_file:
                return jwk_from_pem(pem_file.read())
        except Exception as e:
            raise AnsibleError(f"Error while parsing key file: {e}") from e

    @staticmethod
    def encode_jwt(app_id, jwk, exp=600):
        now = int(time.time())
        payload = {
            "iat": now,
            "exp": now + exp,
            "iss": app_id,
        }
        try:
            return jwt_instance.encode(payload, jwk, alg="RS256")
        except Exception as e:
            raise AnsibleError(f"Error while encoding jwt: {e}") from e


def read_key(path, private_key=None):
    if HAS_PYTHON_JWT:
        return PythonJWT.read_key(path, private_key)
    try:
        if private_key:
            key_bytes = private_key.encode("utf-8")
        else:
            with open(path, "rb") as pem_file:
                key_bytes = pem_file.read()
        return serialization.load_pem_private_key(key_bytes, password=None)
    except Exception as e:
        raise AnsibleError(f"Error while parsing key file: {e}") from e


def encode_jwt(app_id, private_key_obj, exp=600):
    if HAS_PYTHON_JWT:
        return PythonJWT.encode_jwt(app_id, private_key_obj)
    now = int(time.time())
    payload = {
        "iat": now,
        "exp": now + exp,
        "iss": app_id,
    }
    try:
        return jwt.encode(payload, private_key_obj, algorithm="RS256")
    except Exception as e:
        raise AnsibleError(f"Error while encoding jwt: {e}") from e


def post_request(generated_jwt, installation_id, api_base):
    base = api_base.rstrip("/")
    github_url = f"{base}/app/installations/{installation_id}/access_tokens"

    headers = {
        "Authorization": f"Bearer {generated_jwt}",
        "Accept": "application/vnd.github.v3+json",
    }
    try:
        response = open_url(github_url, headers=headers, method="POST")
    except HTTPError as e:
        try:
            error_body = json.loads(e.read().decode())
            display.vvv(f"Error returned: {error_body}")
        except Exception:
            error_body = {}
        if e.code == HTTPStatus.NOT_FOUND:
            raise AnsibleError("Github return error. Please confirm your installation_id value is valid") from e
        elif e.code == HTTPStatus.UNAUTHORIZED:
            raise AnsibleError("Github return error. Please confirm your private key is valid") from e
        raise AnsibleError(f"Unexpected data returned: {e} -- {error_body}") from e
    response_body = response.read()
    try:
        json_data = json.loads(response_body.decode("utf-8"))
    except json.decoder.JSONDecodeError as e:
        raise AnsibleError(f"Error while dencoding JSON respone from github: {e}") from e
    return json_data.get("token")


def get_token(key_path, app_id, installation_id, private_key, github_url, expiry=600):
    jwk = read_key(key_path, private_key)
    generated_jwt = encode_jwt(app_id, jwk, exp=expiry)
    return post_request(generated_jwt, installation_id, github_url)


class LookupModule(LookupBase):
    def run(self, terms, variables=None, **kwargs):
        if not HAS_JWT:
            raise AnsibleError('Python jwt library is required. Please install using "pip install pyjwt"')

        if not HAS_PYTHON_JWT and not HAS_CRYPTOGRAPHY:
            raise AnsibleError(
                'Python cryptography library is required. Please install using "pip install cryptography"'
            )

        self.set_options(var_options=variables, direct=kwargs)
        check_for_no_terms(self, terms=terms, direct=kwargs)

        if not (self.get_option("key_path") or self.get_option("private_key")):
            raise AnsibleOptionsError("One of key_path or private_key is required")
        if self.get_option("key_path") and self.get_option("private_key"):
            raise AnsibleOptionsError("key_path and private_key are mutually exclusive")

        t = get_token(
            self.get_option("key_path"),
            self.get_option("app_id"),
            self.get_option("installation_id"),
            self.get_option("private_key"),
            self.get_option("github_url"),
            self.get_option("token_expiry"),
        )

        return [t]



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2021, Anvitha Jain (@anvitha-jain) <anvjain@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_clone
short_description: Clone schemas
description:
- Clone schemas on Cisco ACI Multi-Site.
- Clones only template objects and not site objects.
- This module can only be used on versions of MSO that are 3.3 or greater.
author:
- Anvitha Jain (@anvitha-jain)
options:
  source_schema:
    description:
    - The name of the source_schema.
    type: str
  destination_schema:
    description:
    - The name of the destination_schema.
    type: str
  state:
    description:
    - Use C(clone) for adding.
    type: str
    choices: [ clone ]
    default: clone
seealso:
- module: cisco.mso.mso_schema
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Clone schema
  cisco.mso.mso_schema_clone:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    source_schema: Source_Schema
    destination_schema: Destination_Schema
    state: clone
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec
from ansible_collections.cisco.mso.plugins.module_utils.constants import NDO_4_UNIQUE_IDENTIFIERS
import json


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        source_schema=dict(type="str"),
        destination_schema=dict(type="str"),
        state=dict(type="str", default="clone", choices=["clone"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "clone", ["destination_schema"]],
        ],
    )

    source_schema = module.params.get("source_schema")
    destination_schema = module.params.get("destination_schema")
    state = module.params.get("state")

    mso = MSOModule(module)

    # Get source schema details
    source_schema_path = "schemas/{0}".format(mso.lookup_schema(source_schema))
    source_schema_obj = mso.query_obj(source_schema_path, displayName=source_schema)

    source_data = source_schema_obj.get("templates")
    source_data = json.loads(json.dumps(source_data).replace("/{0}".format(source_schema_path), ""))
    # certain unique identifiers are present in NDO4.0> source which need to be deleted from source_data prior to POST
    for template in source_data:
        mso.delete_keys_from_dict(template, NDO_4_UNIQUE_IDENTIFIERS)

    path = "schemas"

    # Check if source and destination schema are named differently
    if source_schema == destination_schema:
        mso.fail_json(msg="Source and Destination schema cannot have same names.")
    # Query for existing object(s)
    if destination_schema:
        mso.existing = mso.get_obj(path, displayName=destination_schema)
        if mso.existing:
            mso.fail_json(msg="Schema with the name '{0}' already exists. Please use another name.".format(destination_schema))

    if state == "clone":
        mso.previous = mso.existing
        payload = dict(
            displayName=destination_schema,
            templates=source_data,
        )
        mso.sanitize(payload, collate=True)

        if not mso.existing:
            if module.check_mode:
                mso.existing = mso.proposed
            else:
                mso.existing = mso.request(path, method="POST", data=mso.sent)

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2025, Tim Cragg (@timcragg)
# Copyright: (c) 2025, Shreyas Srish (@shrsr)
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: aci_l4l7_policy_based_redirect_destination
version_added: "2.12.0"
short_description: Manage L4-L7 Policy Based Redirect Destinations (vns:RedirectDest and vns:L1L2RedirectDest)
description:
- Manage L4-L7 Policy Based Redirect Destinations
options:
  tenant:
    description:
    - The name of an existing tenant.
    type: str
    aliases: [ tenant_name ]
  policy:
    description:
    - The name of an existing Policy Based Redirect Policy.
    type: str
    aliases: [ policy_name ]
  description:
    description:
    - The description for redirection.
    type: str
  ip:
    description:
    - The destination IP for redirection.
    - Only used if I(destination_type=l3)
    aliases: [ redirect_ip ]
    type: str
  additional_ip:
    description:
    - The Additional IP Address for the Destination.
    - Only used if I(destination_type=l3)
    type: str
  logical_device:
    description:
    - The destination Logical Device for redirection.
    - Only used if I(destination_type=l1/l2)
    type: str
    aliases: [ logical_dev ]
  concrete_device:
    description:
    - The destination Concrete Device for redirection.
    - Only used if I(destination_type=l1/l2)
    type: str
    aliases: [ concrete_dev ]
  concrete_interface:
    description:
    - The destination Concrete Interface for redirection.
    - Only used if I(destination_type=l1/l2)
    type: str
    aliases: [ concrete_intf ]
  mac:
    description:
    - The destination MAC address for redirection.
    type: str
    aliases: [ redirect_mac ]
  destination_name:
    description:
    - The name for Policy Based Redirect destination.
    type: str
    aliases: [ dest_name ]
  destination_type:
    description:
    - The destination type.
    type: str
    choices: [ l1/l2, l3 ]
    aliases: [ dest_type ]
    default: l3
  pod_id:
    description:
    - The Pod ID to deploy Policy Based Redirect destination on.
    - The APIC defaults to 1 when unset during creation.
    type: int
  health_group:
    description:
    - The Health Group to bind the Policy Based Redirection Destination to.
    - To remove an existing binding from a Health Group, submit a request with I(state=present) and I(health_group="") value.
    type: str
  weight:
    description:
    - The weight of the fault in calculating the health score of an object.
    - The APIC defaults to 1 when unset during creation.
    - Permitted values are in the range of [1, 10].
    type: int
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation
notes:
- The I(tenant), I(device), I(concrete_device), I(concrete_interface) and I(policy) must exist before using this module in your playbook.
  The M(cisco.aci.aci_tenant), M(cisco.aci.aci_l4l7_device), M(cisco.aci.aci_l4l7_concrete_device), M(cisco.aci.aci_l4l7_concrete_interface)
  and M(cisco.aci.aci_l4l7_policy_based_redirect) modules can be used for this.
seealso:
- module: cisco.aci.aci_tenant
- module: cisco.aci.aci_l4l7_device
- module: cisco.aci.aci_l4l7_concrete_device
- module: cisco.aci.aci_l4l7_concrete_interface
- module: cisco.aci.aci_l4l7_policy_based_redirect
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(vns:RedirectDest), B(vns:L1L2RedirectDest)
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Tim Cragg (@timcragg)
- Shreyas Srish (@shrsr)
"""

EXAMPLES = r"""
- name: Add destination to a Policy Based Redirect Policy
  cisco.aci.aci_l4l7_policy_based_redirect_destination:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: my_tenant
    policy: my_pbr_policy
    destination_type: l3
    ip: 192.168.10.1
    mac: AB:CD:EF:12:34:56
    destination_name: redirect_dest
    pod_id: 1
    state: present
  delegate_to: localhost

- name: Query destinations for a Policy Based Redirect Policy
  cisco.aci.aci_l4l7_policy_based_redirect_destination:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: my_tenant
    policy: my_pbr_policy
    state: query
  delegate_to: localhost
  register: query_result

- name: Query destinations for all Policy Based Redirect Policies
  cisco.aci.aci_l4l7_policy_based_redirect_destination:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
  register: query_result

- name: Remove destination from a Policy Based Redirect Policy
  cisco.aci.aci_l4l7_policy_based_redirect_destination:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: my_tenant
    policy: my_pbr_policy
    state: absent
  delegate_to: localhost
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(
        tenant=dict(type="str", aliases=["tenant_name"]),
        policy=dict(type="str", aliases=["policy_name"]),
        ip=dict(type="str", aliases=["redirect_ip"]),
        additional_ip=dict(type="str"),
        mac=dict(type="str", aliases=["redirect_mac"]),
        logical_device=dict(type="str", aliases=["logical_dev"]),
        concrete_device=dict(type="str", aliases=["concrete_dev"]),
        concrete_interface=dict(type="str", aliases=["concrete_intf"]),
        destination_name=dict(type="str", aliases=["dest_name"]),
        destination_type=dict(type="str", aliases=["dest_type"], choices=["l1/l2", "l3"], default="l3"),
        health_group=dict(type="str"),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
        pod_id=dict(type="int"),
        weight=dict(type="int"),
        description=dict(type="str"),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["tenant", "policy"]],
            ["state", "present", ["tenant", "policy"]],
        ],
    )

    aci = ACIModule(module)

    tenant = module.params.get("tenant")
    state = module.params.get("state")
    policy = module.params.get("policy")
    ip = module.params.get("ip")
    additional_ip = module.params.get("additional_ip")
    mac = module.params.get("mac")
    logical_device = module.params.get("logical_device")
    concrete_device = module.params.get("concrete_device")
    concrete_interface = module.params.get("concrete_interface")
    destination_name = module.params.get("destination_name")
    destination_type = module.params.get("destination_type")
    health_group = module.params.get("health_group")
    state = module.params.get("state")
    pod_id = module.params.get("pod_id")
    weight = module.params.get("weight")
    description = module.params.get("description")

    if destination_type == "l1/l2":
        aci_class = "vnsL1L2RedirectDest"
        aci_rn = "L1L2RedirectDest-[{0}]".format(destination_name)
        module_object = destination_name
        target_filter = {"destName": destination_name}
        child_classes = ["vnsRsL1L2RedirectHealthGroup", "vnsRsToCIf"]
        redirect_hg_class = "vnsRsL1L2RedirectHealthGroup"
    else:
        aci_class = "vnsRedirectDest"
        aci_rn = "RedirectDest_ip-[{0}]".format(ip)
        module_object = ip
        target_filter = {"ip": ip}
        child_classes = ["vnsRsRedirectHealthGroup"]
        redirect_hg_class = "vnsRsRedirectHealthGroup"

    aci.construct_url(
        root_class=dict(
            aci_class="fvTenant",
            aci_rn="tn-{0}".format(tenant),
            module_object=tenant,
            target_filter={"name": tenant},
        ),
        subclass_1=dict(
            aci_class="vnsSvcRedirectPol",
            aci_rn="svcCont/svcRedirectPol-{0}".format(policy),
            module_object=policy,
            target_filter={"name": policy},
        ),
        subclass_2=dict(
            aci_class=aci_class,
            aci_rn=aci_rn,
            module_object=module_object,
            target_filter=target_filter,
        ),
        child_classes=child_classes,
    )
    aci.get_existing()

    if state == "present":
        if destination_type == "l1/l2" and additional_ip is not None:
            aci.fail_json(msg="You cannot provide an additional_ip when configuring an l1/l2 destination")
        elif destination_type == "l3" and (logical_device, concrete_device, concrete_interface) != (None, None, None):
            aci.fail_json(msg="You cannot provide a logical_device, concrete_device or concrete_interface when configuring an l3 destination")
        elif destination_type == "l1/l2" and (logical_device, concrete_device, concrete_interface) == (None, None, None):
            aci.fail_json(msg="You must provide a logical_device, concrete_device and concrete_interface when configuring an l1/l2 destination")
        elif destination_type == "l1/l2" and ip is not None:
            aci.fail_json(msg="You cannot provide an ip when configuring an l1/l2 destination")
        child_configs = []
        if destination_type == "l1/l2":
            child_configs = [
                {
                    "vnsRsToCIf": {
                        "attributes": {"tDn": "uni/tn-{0}/lDevVip-{1}/cDev-{2}/cIf-[{3}]".format(tenant, logical_device, concrete_device, concrete_interface)}
                    }
                }
            ]
        if health_group:
            health_group_tdn = "uni/tn-{0}/svcCont/redirectHealthGroup-{1}".format(tenant, health_group)
            child_configs.append({redirect_hg_class: {"attributes": {"tDn": health_group_tdn}}})
        else:
            health_group_tdn = None
        if isinstance(aci.existing, list) and len(aci.existing) > 0:
            for child in aci.existing[0].get(aci_class, {}).get("children", {}):
                if child.get(redirect_hg_class) and child.get(redirect_hg_class).get("attributes").get("tDn") != health_group_tdn:
                    child_configs.append(
                        {
                            redirect_hg_class: {
                                "attributes": {
                                    "dn": child.get(redirect_hg_class).get("attributes").get("dn"),
                                    "status": "deleted",
                                }
                            }
                        }
                    )
        aci.payload(
            aci_class=aci_class,
            class_config=dict(
                ip=ip,
                mac=mac,
                destName=destination_name,
                podId=pod_id,
                ip2=additional_ip,
                weight=weight,
                descr=description,
            ),
            child_configs=child_configs,
        )
        aci.get_diff(aci_class=aci_class)

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2020, Kevin Breit (@kbreit) <kevin.breit@kevinbreit.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {
    'metadata_version': '1.1',
    "status": ['deprecated'],
    'supported_by': 'community'
}

DOCUMENTATION = r'''
---
module: meraki_ms_ospf
short_description: Manage OSPF configuration on MS switches
description:
- Configure OSPF for compatible Meraki MS switches.
deprecated:
  removed_in: '3.0.0'
  why: Updated modules released with increased functionality
  alternative: cisco.meraki.networks_switch_routing_ospf
options:
    state:
        description:
        - Read or edit OSPF settings.
        type: str
        choices: [ present, query ]
        default: present
    net_name:
        description:
        - Name of network containing OSPF configuration.
        type: str
        aliases: [ name, network ]
    net_id:
        description:
        - ID of network containing OSPF configuration.
        type: str
    enabled:
        description:
        - Enable or disable OSPF on the network.
        type: bool
    hello_timer:
        description:
        - Time interval, in seconds, at which hello packets will be sent to OSPF neighbors to maintain connectivity.
        - Value must be between 1 and 255.
        - Default is 10 seconds.
        type: int
    dead_timer:
        description:
        - Time interval to determine when the peer will be declared inactive.
        - Value must be between 1 and 65535.
        type: int
    md5_authentication_enabled:
        description:
        - Whether to enable or disable MD5 authentication.
        type: bool
    md5_authentication_key:
        description:
        - MD5 authentication credentials.
        type: dict
        suboptions:
            id:
                description:
                - MD5 authentication key index.
                - Must be between 1 and 255.
                type: str
            passphrase:
                description:
                - Plain text authentication passphrase
                type: str
    areas:
        description:
        - List of areas in OSPF network.
        type: list
        elements: dict
        suboptions:
            area_id:
                description:
                - OSPF area ID
                type: int
                aliases: [ id ]
            area_name:
                description:
                - Descriptive name of OSPF area.
                type: str
                aliases: [ name ]
            area_type:
                description:
                - OSPF area type.
                choices: [normal, stub, nssa]
                type: str
                aliases: [ type ]
author:
- Kevin Breit (@kbreit)
extends_documentation_fragment: cisco.meraki.meraki
'''

EXAMPLES = r'''
- name: Query OSPF settings
  meraki_ms_ospf:
    auth_key: abc123
    org_name: YourOrg
    net_name: YourNet
    state: query
  delegate_to: localhost

- name: Enable OSPF with check mode
  meraki_ms_ospf:
    auth_key: abc123
    org_name: YourOrg
    net_name: YourNet
    state: present
    enabled: true
    hello_timer: 20
    dead_timer: 60
    areas:
      area_id: 0
      area_name: Backbone
      area_type: normal
      md5_authentication_enabled: false
'''

RETURN = r'''
data:
    description: Information about queried object.
    returned: success
    type: complex
    contains:
        enabled:
            description:
            - Enable or disable OSPF on the network.
            type: bool
        hello_timer_in_seconds:
            description:
            - Time interval, in seconds, at which hello packets will be sent to OSPF neighbors to maintain connectivity.
            type: int
        dead_timer_in_seconds:
            description:
            - Time interval to determine when the peer will be declared inactive.
            type: int
        areas:
            description:
            - List of areas in OSPF network.
            type: complex
            contains:
                area_id:
                    description:
                    - OSPF area ID
                    type: int
                area_name:
                    description:
                    - Descriptive name of OSPF area.
                    type: str
                area_type:
                    description:
                    - OSPF area type.
                    type: str
        md5_authentication_enabled:
            description:
            - Whether to enable or disable MD5 authentication.
            type: bool
        md5_authentication_key:
            description:
            - MD5 authentication credentials.
            type: complex
            contains:
                id:
                    description:
                    - MD5 key index.
                    type: int
                passphrase:
                    description:
                    - Passphrase for MD5 key.
                    type: str
'''

from ansible.module_utils.basic import AnsibleModule, json
from ansible_collections.cisco.meraki.plugins.module_utils.network.meraki.meraki import MerakiModule, meraki_argument_spec


def construct_payload(meraki):
    payload_key_mapping = {'enabled': 'enabled',
                           'hello_timer': 'helloTimerInSeconds',
                           'dead_timer': 'deadTimerInSeconds',
                           'areas': 'areas',
                           'area_id': 'areaId',
                           'area_name': 'areaName',
                           'area_type': 'areaType',
                           'md5_authentication_enabled': 'md5AuthenticationEnabled',
                           'md5_authentication_key': 'md5AuthenticationKey',
                           'id': 'id',
                           'passphrase': 'passphrase',
                           }
    payload = {}

    # This may need to be reworked to avoid overwiting
    for snake, camel in payload_key_mapping.items():
        try:
            if meraki.params[snake] is not None:
                payload[camel] = meraki.params[snake]
                if snake == 'areas':
                    if meraki.params['areas'] is not None and len(meraki.params['areas']) > 0:
                        payload['areas'] = []
                        for area in meraki.params['areas']:
                            area_settings = {'areaName': area['area_name'],
                                             'areaId': area['area_id'],
                                             'areaType': area['area_type'],
                                             }
                            payload['areas'].append(area_settings)
                # TODO: Does this code below have a purpose?
                # elif snake == 'md5_authentication_key':
                #     if meraki.params['md5_authentication_key'] is not None:
                #         md5_settings = {'id': meraki.params['md5_authentication_key']['id'],
                #                         'passphrase': meraki.params['md5_authentication_key']['passphrase'],
                #                         }
        except KeyError:
            pass

    return payload


def main():
    # define the available arguments/parameters that a user can pass to
    # the module

    areas_arg_spec = dict(area_id=dict(type='int', aliases=['id']),
                          area_name=dict(type='str', aliases=['name']),
                          area_type=dict(type='str', aliases=['type'], choices=['normal', 'stub', 'nssa']),
                          )

    md5_auth_arg_spec = dict(id=dict(type='str'),
                             passphrase=dict(type='str', no_log=True),
                             )

    argument_spec = meraki_argument_spec()
    argument_spec.update(state=dict(type='str', choices=['present', 'query'], default='present'),
                         net_id=dict(type='str'),
                         net_name=dict(type='str', aliases=['name', 'network']),
                         enabled=dict(type='bool'),
                         hello_timer=dict(type='int'),
                         dead_timer=dict(type='int'),
                         areas=dict(type='list', default=None, elements='dict', options=areas_arg_spec),
                         md5_authentication_enabled=dict(type='bool'),
                         md5_authentication_key=dict(type='dict', default=None, options=md5_auth_arg_spec, no_log=True),
                         )

    # the AnsibleModule object will be our abstraction working with Ansible
    # this includes instantiation, a couple of common attr would be the
    # args/params passed to the execution, as well as if the module
    # supports check mode
    module = AnsibleModule(argument_spec=argument_spec,
                           supports_check_mode=True,
                           )
    meraki = MerakiModule(module, function='ms_ospf')

    meraki.params['follow_redirects'] = 'all'

    query_urls = {'ms_ospf': '/networks/{net_id}/switch/routing/ospf'}
    update_urls = {'ms_ospf': '/networks/{net_id}/switch/routing/ospf'}

    meraki.url_catalog['get_all'].update(query_urls)
    meraki.url_catalog['update'] = update_urls

    payload = None

    # execute checks for argument completeness

    if meraki.params['dead_timer'] is not None:
        if meraki.params['dead_timer'] < 1 or meraki.params['dead_timer'] > 65535:
            meraki.fail_json(msg='dead_timer must be between 1 and 65535')
    if meraki.params['hello_timer'] is not None:
        if meraki.params['hello_timer'] < 1 or meraki.params['hello_timer'] > 255:
            meraki.fail_json(msg='hello_timer must be between 1 and 65535')
    if meraki.params['md5_authentication_enabled'] is False:
        if meraki.params['md5_authentication_key'] is not None:
            meraki.fail_json(msg='md5_authentication_key must not be configured when md5_authentication_enabled is false')

    # manipulate or modify the state as needed (this is going to be the
    # part where your module will do what it needs to do)

    org_id = meraki.params['org_id']
    if not org_id:
        org_id = meraki.get_org_id(meraki.params['org_name'])
    net_id = meraki.params['net_id']
    if net_id is None and meraki.params['net_name']:
        nets = meraki.get_nets(org_id=org_id)
        net_id = meraki.get_net_id(net_name=meraki.params['net_name'], data=nets)
    if meraki.params['state'] == 'query':
        path = meraki.construct_path('get_all', net_id=net_id)
        response = meraki.request(path, method='GET')
        meraki.result['data'] = response
        meraki.exit_json(**meraki.result)
    elif meraki.params['state'] == 'present':
        original = meraki.request(meraki.construct_path('get_all', net_id=net_id), method='GET')
        payload = construct_payload(meraki)
        if meraki.is_update_required(original, payload) is True:
            if meraki.check_mode is True:
                meraki.result['data'] = payload
                meraki.result['changed'] = True
                meraki.exit_json(**meraki.result)
            path = meraki.construct_path('update', net_id=net_id)
            response = meraki.request(path, method='PUT', payload=json.dumps(payload))
            if 'md5_authentication_key' in response:
                response['md5_authentication_key']['passphrase'] = 'VALUE_SPECIFIED_IN_NO_LOG_PARAMETER'
            meraki.result['data'] = response
            meraki.result['changed'] = True
            meraki.exit_json(**meraki.result)
        else:
            if 'md5_authentication_key' in original:
                original['md5_authentication_key']['passphrase'] = 'VALUE_SPECIFIED_IN_NO_LOG_PARAMETER'
            meraki.result['data'] = original
            meraki.exit_json(**meraki.result)

    # in the event of a successful module execution, you will want to
    # simple AnsibleModule.exit_json(), passing the key/value results
    meraki.exit_json(**meraki.result)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright (c) 2018 Red Hat, Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import copy
import typing

if typing.TYPE_CHECKING:
    from typing import List
    from typing import Optional

from ansible.module_utils.basic import to_text

from ansible_collections.amazon.aws.plugins.module_utils.transformation import boto3_resource_to_ansible_dict


def normalize_s3_bucket_versioning(versioning_status: Optional[dict]) -> Optional[dict]:
    """
    Normalize S3 bucket versioning status to Ansible dictionary format.

    Parameters:
        versioning_status (dict): The versioning status returned from boto3.

    Returns:
        Normalized versioning status dictionary with both legacy and current formats.
    """
    if not versioning_status:
        return versioning_status
    versioning_result = typing.cast(dict, boto3_resource_to_ansible_dict(versioning_status))
    # Original s3_bucket format, no longer advertised but not officially deprecated
    versioning_result["Versioning"] = versioning_status.get("Status", "Disabled")
    versioning_result["MfaDelete"] = versioning_status.get("MFADelete", "Disabled")
    # Original s3_bucket_info format, no longer advertised but not officially deprecated
    versioning_result["Status"] = versioning_status.get("Status", "Disabled")
    versioning_result["MFADelete"] = versioning_status.get("MFADelete", "Disabled")
    return versioning_result


def normalize_s3_bucket_public_access(public_access_status: Optional[dict]) -> Optional[dict]:
    """
    Normalize S3 bucket public access block configuration to Ansible dictionary format.

    Parameters:
        public_access_status (dict): The public access block configuration from boto3.

    Returns:
        Normalized public access configuration dictionary.
    """
    if not public_access_status:
        return public_access_status
    public_access_result = typing.cast(dict, boto3_resource_to_ansible_dict(public_access_status))
    public_access_result["PublicAccessBlockConfiguration"] = copy.deepcopy(public_access_status)
    public_access_result.update(public_access_status)
    return public_access_result


def normalize_s3_bucket_acls(acls: Optional[dict]) -> Optional[dict]:
    """
    Normalize S3 bucket ACLs to Ansible dictionary format.

    Parameters:
        acls (dict): The ACL configuration returned from boto3.

    Returns:
        Normalized ACL grants dictionary.
    """
    if not acls:
        return acls
    acls_result = typing.cast(dict, boto3_resource_to_ansible_dict(acls))
    return typing.cast(dict, acls_result["grants"])


def normalize_s3_bucket_logging(logging_config: Optional[dict]) -> Optional[dict]:
    """
    Normalize S3 bucket logging configuration to Ansible dictionary format.

    Parameters:
        logging_config (dict): The logging configuration returned from boto3.

    Returns:
        Normalized logging configuration dictionary or None if logging is disabled.
    """
    if not logging_config or "LoggingEnabled" not in logging_config:
        return None

    result = boto3_resource_to_ansible_dict(logging_config.get("LoggingEnabled"), force_tags=False)
    # Ensure target_prefix is always present, even if not in source
    if "target_prefix" not in result:
        result["target_prefix"] = ""
    return result


def _grantee_is_owner(grant: dict, owner_id: str) -> bool:
    return grant.get("Grantee", {}).get("ID") == owner_id


def _grantee_is_public(grant: dict) -> bool:
    return grant.get("Grantee", {}).get("URI") == "http://acs.amazonaws.com/groups/global/AllUsers"


def _grantee_is_authenticated(grant: dict) -> bool:
    return grant.get("Grantee", {}).get("URI") == "http://acs.amazonaws.com/groups/global/AuthenticatedUsers"


def _acl_permissions(grants: Optional[List[dict]]) -> List[str]:
    if not grants:
        return []
    return [grant.get("Permission") for grant in grants if grant]


def s3_acl_to_name(acl: Optional[dict]) -> Optional[str]:
    """
    Convert S3 ACL grant structure to a named canned ACL template.

    Attempts to match the ACL grants to standard AWS canned ACLs by analyzing
    the owner, authenticated user, and public grants.

    Parameters:
        acl (dict): The ACL structure with 'Grants' and 'Owner' keys.

    Returns:
        The name of the matching canned ACL ('private', 'public-read',
        'public-read-write', 'authenticated-read'), or None if no match is found.
    """
    if not acl:
        return None

    try:
        grants = acl["Grants"]
        owner_id = acl["Owner"]["ID"]
        owner_acl = [grant for grant in grants if _grantee_is_owner(grant, owner_id)]
        auth_acl = [grant for grant in grants if _grantee_is_authenticated(grant)]
        public_acl = [grant for grant in grants if _grantee_is_public(grant)]

        if len(grants) > (len(owner_acl) + len(auth_acl) + len(public_acl)):
            raise ValueError("Unrecognised Grantee")
        if public_acl and auth_acl:
            raise ValueError("Public ACLs and Authenticated User ACLs are only used alone in templated ACL")

        if ["FULL_CONTROL"] != _acl_permissions(owner_acl):
            raise ValueError("Owner doesn't have full control")
        if len(grants) == 1:
            return "private"

        if auth_acl:
            if ["READ"] == _acl_permissions(auth_acl):
                return "authenticated-read"
            raise ValueError("Authenticated User ACLs don't match templated ACL")

        permissions = sorted(_acl_permissions(public_acl))
        if permissions == ["READ"]:
            return "public-read"
        if permissions == ["READ", "WRITE"]:
            return "public-read-write"

        raise ValueError("Public ACLs don't match templated ACL")

    except (KeyError, IndexError, ValueError):
        return None


def merge_tags(current_tags: dict, new_tags: dict, purge_tags: bool = True) -> dict:
    """
    Compare and merge two dicts of tags.

    :param current_tags: The current tags
    :param new_tags: The tags passed as a parameter
    :param purge_tags: Whether to remove current tags that aren't in new_tags
    :return: updated_tags: The updated dictionary of tags
    """

    if new_tags is None:
        return current_tags

    updated_tags = copy.deepcopy(current_tags) if not purge_tags else {}
    # Tags are always returned as text
    new_tags = dict((to_text(k), to_text(v)) for k, v in new_tags.items())
    updated_tags.update(new_tags)

    return updated_tags



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""
ntc_templates parser

This is the ntc_templates parser for use with the cli_parse module and action plugin.
https://github.com/networktocode/ntc-templates

"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
    author: Bradley Thornton (@cidrblock)
    name: ntc_templates
    short_description: Define configurable options for C(ntc_templates) sub-plugin of C(cli_parse) module
    description:
    - This plugin documentation provides the configurable options that can be passed
      to the I(ansible.utils.cli_parse) plugins when I(ansible.netcommon.ntc_templates) is used as a value for
      I(name) option.
    version_added: 1.0.0
"""

EXAMPLES = r"""
- name: "Run command and parse with ntc_templates"
  ansible.utils.cli_parse:
    command: "show interface"
    parser:
      name: ansible.netcommon.ntc_templates
  register: nxos_ntc_templates_command

- name: "Pass text and command"
  ansible.utils.cli_parse:
    text: "{{ nxos_ntc_templates_command['stdout'] }}"
    parser:
      name: ansible.netcommon.ntc_templates
      command: show interface
  register: nxos_ntc_templates_text
"""

from ansible.module_utils.basic import missing_required_lib
from ansible.module_utils.common.text.converters import to_native
from ansible_collections.ansible.utils.plugins.plugin_utils.base.cli_parser import CliParserBase


try:
    from ntc_templates.parse import parse_output

    HAS_NTC = True
except ImportError:
    HAS_NTC = False

ANSIBLE_NETWORK_OS = {
    "ios": "cisco_ios",
    "iosxr": "cisco_xr",
    "nxos": "cisco_nxos",
    "asa": "cisco_asa",
    "eos": "arista_eos",
    "junos": "juniper_junos",
}


class CliParser(CliParserBase):
    """The ntc_templates parser class
    Convert raw text to structured data using textfsm and predefined templates in
    the ntc-templates python package
    """

    DEFAULT_TEMPLATE_EXTENSION = None
    PROVIDE_TEMPLATE_CONTENTS = False

    def _check_reqs(self):
        """Check the prerequisites for the ntc template parser

        :return: A dict with errors or a network_os and command
        :rtype: dict
        """
        errors = []

        if not HAS_NTC:
            errors.append(missing_required_lib("ntc-templates"))

        network_os = self._task_args.get("parser").get("os")
        if network_os:
            self._debug("OS set to {os} using task args".format(os=network_os))
        if not network_os:
            ano = self._task_vars.get("ansible_network_os", "").split(".")[-1]
            network_os = ANSIBLE_NETWORK_OS[ano]
            self._debug("OS set to {os} using ansible_network_os".format(os=network_os))
        if not network_os:
            errors.append("Either 'parser/os' needs to be specified or 'ansible_network_os' set.")
        command = self._task_args.get("parser").get("command")
        if not command:
            errors.append("'command' needs to be specified.")

        if errors:
            return {"errors": errors}
        return {"network_os": network_os, "command": command}

    def parse(self, *_args, **_kwargs):
        """Std entry point for a cli_parse parse execution

        :return: Errors or parsed text as structured data
        :rtype: dict

        :example:

        The parse function of a parser should return a dict:
        {"errors": [a list of errors]}
        or
        {"parsed": obj}
        """
        cli_output = self._task_args.get("text")
        res = self._check_reqs()
        if res.get("errors"):
            return {"errors": res.get("errors")}
        platform = res["network_os"]
        command = res["command"]
        try:
            parsed = parse_output(platform=platform, command=command, data=cli_output)
            return {"parsed": parsed}
        except Exception as exc:
            return {"errors": [to_native(exc)]}



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: iam_access_key
version_added: 2.1.0
version_added_collection: community.aws
short_description: Manage AWS IAM User access keys
description:
  - Manage AWS IAM user access keys.
author:
  - Mark Chappell (@tremble)
options:
  user_name:
    description:
      - The name of the IAM User to which the key belongs.
    required: true
    type: str
    aliases: ['username']
  id:
    description:
      - The ID of the access key.
      - Required when O(state=absent).
      - Mutually exclusive with O(rotate_keys).
    required: false
    type: str
  state:
    description:
      - Create or remove the access key.
      - When O(state=present) and O(id) is not defined a new key will be created.
    required: false
    type: str
    default: 'present'
    choices: [ 'present', 'absent' ]
  active:
    description:
      - Whether the key should be enabled or disabled.
      - Defaults to V(true) when creating a new key.
    required: false
    type: bool
    aliases: ['enabled']
  rotate_keys:
    description:
      - When there are already 2 access keys attached to the IAM user the oldest
        key will be removed and a new key created.
      - Ignored if O(state=absent)
      - Mutually exclusive with O(id).
    required: false
    type: bool
    default: false
notes:
  - For security reasons, this module should be used with B(no_log=true) and (register) functionalities
    when creating new access key.
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

- name: Create a new access key
  amazon.aws.iam_access_key:
    user_name: example_user
    state: present
  no_log: true

- name: Delete the access_key
  amazon.aws.iam_access_key:
    user_name: example_user
    id: AKIA1EXAMPLE1EXAMPLE
    state: absent
"""

RETURN = r"""
access_key:
    description: A dictionary containing all the access key information.
    returned: When the key exists.
    type: complex
    contains:
        access_key_id:
            description: The ID for the access key.
            returned: success
            type: str
            sample: AKIA1EXAMPLE1EXAMPLE
        create_date:
            description: The date and time, in ISO 8601 date-time format, when the access key was created.
            returned: success
            type: str
            sample: "2021-10-09T13:25:42+00:00"
        user_name:
            description: The name of the IAM user to which the key is attached.
            returned: success
            type: str
            sample: example_user
        status:
            description:
              - The status of the key.
              - C(Active) means it can be used.
              - C(Inactive) means it can not be used.
            returned: success
            type: str
            sample: Inactive
secret_access_key:
    description:
      - The secret access key.
      - A secret access key is the equivalent of a password which can not be changed and as such should be considered sensitive data.
      - Secret access keys can only be accessed at creation time.
    returned: When a new key is created.
    type: str
    sample: example/Example+EXAMPLE+example/Example
deleted_access_key_id:
    description:
      - The access key deleted during rotation.
    returned: When a key was deleted during the rotation of access keys
    type: str
    sample: AKIA1EXAMPLE1EXAMPLE
"""

from ansible_collections.amazon.aws.plugins.module_utils.iam import AnsibleIAMError
from ansible_collections.amazon.aws.plugins.module_utils.iam import IAMErrorHandler
from ansible_collections.amazon.aws.plugins.module_utils.iam import get_iam_access_keys
from ansible_collections.amazon.aws.plugins.module_utils.iam import normalize_iam_access_key
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.transformation import scrub_none_parameters


@IAMErrorHandler.deletion_error_handler("Failed to delete access key for user")
def delete_access_key(access_keys, user, access_key_id):
    if not access_key_id:
        return False
    if access_key_id not in [k["access_key_id"] for k in access_keys]:
        return False

    if module.check_mode:
        return True

    client.delete_access_key(aws_retry=True, UserName=user, AccessKeyId=access_key_id)
    return True


@IAMErrorHandler.common_error_handler("Failed to update access key for user")
def update_access_key_state(access_keys, user, access_key_id, enabled):
    keys = {k["access_key_id"]: k for k in access_keys}

    if access_key_id not in keys:
        raise AnsibleIAMError(message=f'Access key "{access_key_id}" not found attached to User "{user}"')

    if enabled is None:
        return False

    access_key = keys.get(access_key_id)

    desired_status = "Active" if enabled else "Inactive"
    if access_key.get("status") == desired_status:
        return False

    if module.check_mode:
        return True

    client.update_access_key(aws_retry=True, UserName=user, AccessKeyId=access_key_id, Status=desired_status)
    return True


@IAMErrorHandler.common_error_handler("Failed to create access key for user")
def create_access_key(access_keys, user, rotate_keys, enabled):
    changed = False
    oldest_key = False

    if len(access_keys) > 1 and rotate_keys:
        oldest_key = access_keys[0].get("access_key_id")
        changed |= delete_access_key(access_keys, user, oldest_key)

    if module.check_mode:
        if oldest_key:
            return dict(deleted_access_key=oldest_key)
        return dict()

    results = client.create_access_key(aws_retry=True, UserName=user)
    access_key = normalize_iam_access_key(results.get("AccessKey"))

    # Update settings which can't be managed on creation
    if enabled is False:
        access_key_id = access_key["access_key_id"]
        update_access_key_state([access_key], user, access_key_id, enabled)
        access_key["status"] = "Inactive"

    if oldest_key:
        access_key["deleted_access_key"] = oldest_key

    return access_key


def update_access_key(access_keys, user, access_key_id, enabled):
    changed = update_access_key_state(access_keys, user, access_key_id, enabled)
    access_keys = get_iam_access_keys(client, user)
    keys = {k["access_key_id"]: k for k in access_keys}
    return changed, {"access_key": keys.get(access_key_id, None)}


def main():
    global module
    global client

    argument_spec = dict(
        user_name=dict(required=True, type="str", aliases=["username"]),
        id=dict(required=False, type="str"),
        state=dict(required=False, choices=["present", "absent"], default="present"),
        active=dict(required=False, type="bool", aliases=["enabled"]),
        rotate_keys=dict(required=False, type="bool", default=False),
    )

    required_if = [
        ["state", "absent", ("id",)],
    ]
    mutually_exclusive = [
        ["rotate_keys", "id"],
    ]

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=required_if,
        mutually_exclusive=mutually_exclusive,
    )

    client = module.client("iam", retry_decorator=AWSRetry.jittered_backoff())

    state = module.params.get("state")
    user = module.params.get("user_name")
    access_key_id = module.params.get("id")
    rotate_keys = module.params.get("rotate_keys")
    enabled = module.params.get("active")

    access_keys = get_iam_access_keys(client, user)
    results = dict()

    try:
        if state == "absent":
            changed = delete_access_key(access_keys, user, access_key_id)
            module.exit_json(changed=changed)

        if access_key_id:
            changed, results = update_access_key(access_keys, user, access_key_id, enabled)
        else:
            secret_key = create_access_key(access_keys, user, rotate_keys, enabled)
            changed = True
            results = {
                "access_key_id": secret_key.get("access_key_id", None),
                "secret_access_key": secret_key.pop("secret_access_key", None),
                "deleted_access_key_id": secret_key.pop("deleted_access_key", None),
                "access_key": secret_key or None,
            }
            results = scrub_none_parameters(results)

        module.exit_json(changed=changed, **results)
    except AnsibleIAMError as e:
        module.fail_json_aws_error(e)


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, division, print_function


__metaclass__ = type


DOCUMENTATION = """
module: eos_user
author: Peter Sprygada (@privateip)
short_description: Manage the collection of local users on EOS devices
description:
- This module provides declarative management of the local usernames configured on
  Arista EOS devices.  It allows playbooks to manage either individual usernames or
  the collection of usernames in the current running config.  It also supports purging
  usernames from the configuration that are not explicitly defined.
version_added: 1.0.0
notes:
- Tested against Arista EOS 4.24.6F
options:
  aggregate:
    description:
    - The set of username objects to be configured on the remote Arista EOS device.  The
      list entries can either be the username or a hash of username and properties.  This
      argument is mutually exclusive with the C(username) argument.
    aliases:
    - users
    - collection
    type: list
    elements: dict
    suboptions:
      name:
        description:
        - The username to be configured on the remote Arista EOS device.  This argument
          accepts a stringv value and is mutually exclusive with the C(aggregate) argument.
        type: str
      configured_password:
        description:
        - The password to be configured on the remote Arista EOS device. The password
          needs to be provided in clear and it will be encrypted on the device.
        type: str
      update_password:
        description:
        - Since passwords are encrypted in the device running config, this argument will
          instruct the module when to change the password.  When set to C(always), the
          password will always be updated in the device and when set to C(on_create) the
          password will be updated only if the username is created.
        type: str
        choices:
        - on_create
        - always
      privilege:
        description:
        - The C(privilege) argument configures the privilege level of the user when logged
          into the system.  This argument accepts integer values in the range of 1 to
          15.
        type: int
      role:
        description:
        - Configures the role for the username in the device running configuration.  The
          argument accepts a string value defining the role name.  This argument does
          not check if the role has been configured on the device.
        type: str
      sshkey:
        description:
        - Specifies the SSH public key to configure for the given username.  This argument
          accepts a valid SSH key value.
        type: str
      nopassword:
        description:
        - Defines the username without assigning a password.  This will allow the user
          to login to the system without being authenticated by a password.
        type: bool
      state:
        description:
        - Configures the state of the username definition as it relates to the device
          operational configuration.  When set to I(present), the username(s) should be
          configured in the device active configuration and when set to I(absent) the
          username(s) should not be in the device active configuration
        type: str
        choices:
        - present
        - absent
  name:
    description:
    - The username to be configured on the remote Arista EOS device.  This argument
      accepts a stringv value and is mutually exclusive with the C(aggregate) argument.
    type: str
  configured_password:
    description:
    - The password to be configured on the remote Arista EOS device. The password
      needs to be provided in clear and it will be encrypted on the device.
    type: str
  update_password:
    description:
    - Since passwords are encrypted in the device running config, this argument will
      instruct the module when to change the password.  When set to C(always), the
      password will always be updated in the device and when set to C(on_create) the
      password will be updated only if the username is created.
    default: always
    type: str
    choices:
    - on_create
    - always
  privilege:
    description:
    - The C(privilege) argument configures the privilege level of the user when logged
      into the system.  This argument accepts integer values in the range of 1 to
      15.
    type: int
  role:
    description:
    - Configures the role for the username in the device running configuration.  The
      argument accepts a string value defining the role name.  This argument does
      not check if the role has been configured on the device.
    type: str
  sshkey:
    description:
    - Specifies the SSH public key to configure for the given username.  This argument
      accepts a valid SSH key value.
    type: str
  nopassword:
    description:
    - Defines the username without assigning a password.  This will allow the user
      to login to the system without being authenticated by a password.
    type: bool
  purge:
    description:
    - Instructs the module to consider the resource definition absolute.  It will
      remove any previously configured usernames on the device with the exception
      of the `admin` user which cannot be deleted per EOS constraints.
    type: bool
    default: false
  state:
    description:
    - Configures the state of the username definition as it relates to the device
      operational configuration.  When set to I(present), the username(s) should be
      configured in the device active configuration and when set to I(absent) the
      username(s) should not be in the device active configuration
    type: str
    default: present
    choices:
    - present
    - absent
"""

EXAMPLES = """
- name: create a new user
  arista.eos.eos_user:
    name: ansible
    sshkey: "{{ lookup('file', '~/.ssh/id_rsa.pub') }}"
    state: present

- name: remove all users except admin
  arista.eos.eos_user:
    purge: true

- name: set multiple users to privilege level 15
  arista.eos.eos_user:
    aggregate:
      - name: netop
      - name: netend
    privilege: 15
    state: present

- name: Change Password for User netop
  arista.eos.eos_user:
    username: netop
    configured_password: '{{ new_password }}'
    update_password: always
    state: present
"""

RETURN = """
commands:
  description: The list of configuration mode commands to send to the device
  returned: always
  type: list
  sample:
    - name ansible secret password
    - name admin secret admin
session_name:
  description: The EOS config session name used to load the configuration
  returned: when changed is True
  type: str
  sample: ansible_1479315771
"""

import re

from copy import deepcopy
from functools import partial

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    remove_default_spec,
)

from ansible_collections.arista.eos.plugins.module_utils.network.eos.eos import (
    get_config,
    load_config,
    run_commands,
)


def validate_privilege(value, module):
    if not 1 <= value <= 15:
        module.fail_json(
            msg="privilege must be between 1 and 15, got %s" % value,
        )


def get_os_version(module):
    os_version = "4.20.10"
    response = run_commands(
        module,
        'show version | grep "Software image version"',
    )
    version_match = re.search(
        r"Software image version:\s+([\d\.]+)",
        response[0],
        re.M,
    )
    if version_match:
        v = version_match.group(1).split(".")
        os_version = tuple(int(digit) for digit in v)
    return os_version


def map_obj_to_commands(updates, module):
    commands = list()
    update_password = module.params["update_password"]

    for update in updates:
        want, have = update

        def needs_update(x):
            return want.get(x) and (want.get(x) != have.get(x))

        def add(x):
            return commands.append("username %s %s" % (want["name"], x))

        if want["state"] == "absent":
            commands.append("no username %s" % want["name"])
            continue

        if needs_update("configured_password"):
            if update_password == "always" or not have:
                add("secret %s" % want["configured_password"])

        if needs_update("role"):
            add("role %s" % want["role"])

        if needs_update("privilege"):
            add("privilege %s" % want["privilege"])

        if needs_update("sshkey"):
            ver = get_os_version(module)
            # compare against image version 4.20.10
            if ver > (4, 20, 10):
                add("ssh-key %s" % want["sshkey"])
            else:
                add("sshkey %s" % want["sshkey"])

        if needs_update("nopassword"):
            if want["nopassword"]:
                add("nopassword")
            else:
                add("no username %s nopassword" % want["name"])

        if want.get("state") == "present" and want.get("name"):
            value = [
                want.get("configured_password"),
                want.get("nopassword"),
                want.get("sshkey"),
            ]
            if all(v is None for v in value) is True:
                module.fail_json(
                    msg="configured_password, sshkey or nopassword should be provided",
                )
    return commands


def parse_role(data):
    match = re.search(r"role (\S+)", data, re.M)
    if match:
        return match.group(1)


def parse_sshkey(data):
    match = re.search(r"sshkey|ssh-key (.+)$", data, re.M)
    if match:
        return match.group(1)


def parse_privilege(data):
    match = re.search(r"privilege (\S+)", data, re.M)
    if match:
        return int(match.group(1))


def map_config_to_obj(module):
    data = get_config(module, flags=["section username"])

    match = re.findall(r"^username (\S+)", data, re.M)
    if not match:
        return list()

    instances = list()

    for user in set(match):
        regex = r"username %s .+$" % user
        cfg = re.findall(regex, data, re.M)
        cfg = "\n".join(cfg)
        obj = {
            "name": user,
            "state": "present",
            "nopassword": "nopassword" in cfg,
            "configured_password": None,
            "sshkey": parse_sshkey(cfg),
            "privilege": parse_privilege(cfg),
            "role": parse_role(cfg),
        }
        instances.append(obj)

    return instances


def get_param_value(key, item, module):
    # if key doesn't exist in the item, get it from module.params
    if not item.get(key):
        value = module.params[key]

    # if key does exist, do a type check on it to validate it
    else:
        value_type = module.argument_spec[key].get("type", "str")
        type_checker = module._CHECK_ARGUMENT_TYPES_DISPATCHER[value_type]
        type_checker(item[key])
        value = item[key]

    # validate the param value (if validator func exists)
    validator = globals().get("validate_%s" % key)
    if all((value, validator)):
        validator(value, module)

    return value


def map_params_to_obj(module):
    aggregate = module.params["aggregate"]
    if not aggregate:
        if not module.params["name"] and module.params["purge"]:
            return list()
        elif not module.params["name"]:
            module.fail_json(msg="name is required")
        else:
            collection = [{"name": module.params["name"]}]
    else:
        collection = list()
        for item in aggregate:
            if not isinstance(item, dict):
                collection.append({"name": item})
            elif "name" not in item:
                module.fail_json(msg="name is required")
            else:
                collection.append(item)

    objects = list()

    for item in collection:
        get_value = partial(get_param_value, item=item, module=module)
        item["configured_password"] = get_value("configured_password")
        item["nopassword"] = get_value("nopassword")
        item["privilege"] = get_value("privilege")
        item["role"] = get_value("role")
        item["sshkey"] = get_value("sshkey")
        item["state"] = get_value("state")
        objects.append(item)

    return objects


def update_objects(want, have):
    updates = list()
    for entry in want:
        if "name" in entry:
            item = next((i for i in have if i["name"] == entry["name"]), None)
        if all((item is None, entry["state"] == "present")):
            updates.append((entry, {}))
        elif item:
            for key, value in entry.items():
                if value and value != item[key]:
                    updates.append((entry, item))
    return updates


def main():
    """main entry point for module execution"""
    element_spec = dict(
        name=dict(),
        configured_password=dict(no_log=True),
        nopassword=dict(type="bool"),
        update_password=dict(
            default="always",
            choices=["on_create", "always"],
        ),
        privilege=dict(type="int"),
        role=dict(),
        sshkey=dict(no_log=True),
        state=dict(default="present", choices=["present", "absent"]),
    )

    aggregate_spec = deepcopy(element_spec)

    # remove default in aggregate spec, to handle common arguments
    remove_default_spec(aggregate_spec)

    argument_spec = dict(
        aggregate=dict(
            type="list",
            elements="dict",
            options=aggregate_spec,
            aliases=["collection", "users"],
        ),
        purge=dict(type="bool", default=False),
    )

    argument_spec.update(element_spec)
    mutually_exclusive = [("name", "aggregate")]

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=mutually_exclusive,
        supports_check_mode=True,
    )

    warnings = list()

    result = {"changed": False}
    if warnings:
        result["warnings"] = warnings

    want = map_params_to_obj(module)
    have = map_config_to_obj(module)

    commands = map_obj_to_commands(update_objects(want, have), module)

    if module.params["purge"]:
        want_users = [x["name"] for x in want]
        have_users = [x["name"] for x in have]
        for item in set(have_users).difference(want_users):
            if item != "admin":
                commands.append("no username %s" % item)

    result["commands"] = commands

    # the eos cli prevents this by rule so capture it and display
    # a nice failure message
    if "no username admin" in commands:
        module.fail_json(msg="cannot delete the `admin` account")

    if commands:
        commit = not module.check_mode
        response = load_config(module, commands, commit=commit)
        if response.get("diff") and module._diff:
            result["diff"] = {"prepared": response.get("diff")}
        result["session_name"] = response.get("session")
        result["changed"] = True

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2023, Ondrej Zvara (ozvara1@gmail.com)
# Based on code:
# Copyright (c) 2021, Lennert Mertens (lennert@nubera.be)
# Copyright (c) 2021, Werner Dijkerman (ikben@werner-dijkerman.nl)
# Copyright (c) 2015, Werner Dijkerman (ikben@werner-dijkerman.nl)
# Copyright (c) 2019, Guillaume Martinez (lunik@tiwabbit.fr)
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
module: gitlab_issue
short_description: Create, update, or delete GitLab issues
version_added: '8.1.0'
description:
  - Creates an issue if it does not exist.
  - When an issue does exist, it is updated if the provided parameters are different.
  - When an issue does exist and O(state=absent), the issue is deleted.
  - When multiple issues are detected, the task fails.
  - Existing issues are matched based on O(title) and O(state_filter) filters.
author:
  - zvaraondrej (@zvaraondrej)
requirements:
  - python-gitlab >= 2.3.0
extends_documentation_fragment:
  - community.general._auth_basic
  - community.general._gitlab
  - community.general._attributes

attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  assignee_ids:
    description:
      - A list of assignee usernames omitting V(@) character.
      - Set to an empty array to unassign all assignees.
    type: list
    elements: str
  description:
    description:
      - A description of the issue.
      - Gets overridden by a content of file specified at O(description_path), if found.
    type: str
  description_path:
    description:
      - A path of file containing issue's description.
      - Accepts MarkDown formatted files.
    type: path
  issue_type:
    description:
      - Type of the issue.
    default: issue
    type: str
    choices: ["issue", "incident", "test_case"]
  labels:
    description:
      - A list of label names.
      - Set to an empty array to remove all labels.
    type: list
    elements: str
  milestone_search:
    description:
      - The name of the milestone.
      - Set to empty string to unassign milestone.
    type: str
  milestone_group_id:
    description:
      - The path or numeric ID of the group hosting desired milestone.
    type: str
  project:
    description:
      - The path or name of the project.
    required: true
    type: str
  state:
    description:
      - Create or delete issue.
    default: present
    type: str
    choices: ["present", "absent"]
  state_filter:
    description:
      - Filter specifying state of issues while searching.
    type: str
    choices: ["opened", "closed"]
    default: opened
  title:
    description:
      - A title for the issue. The title is used as a unique identifier to ensure idempotency.
    type: str
    required: true
"""


EXAMPLES = r"""
- name: Create Issue
  community.general.gitlab_issue:
    api_url: https://gitlab.com
    api_token: secret_access_token
    project: "group1/project1"
    title: "Ansible demo Issue"
    description: "Demo Issue description"
    labels:
      - Ansible
      - Demo
    assignee_ids:
      - testassignee
    state_filter: "opened"
    state: present

- name: Delete Issue
  community.general.gitlab_issue:
    api_url: https://gitlab.com
    api_token: secret_access_token
    project: "group1/project1"
    title: "Ansible demo Issue"
    state_filter: "opened"
    state: absent
"""

RETURN = r"""
msg:
  description: Success or failure message.
  returned: always
  type: str
  sample: "Success"

issue:
  description: API object.
  returned: success
  type: dict
"""

from ansible.module_utils.api import basic_auth_argument_spec
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.general.plugins.module_utils._gitlab import (
    auth_argument_spec,
    find_group,
    find_project,
    gitlab,
    gitlab_authentication,
)


class GitlabIssue:
    def __init__(self, module, project, gitlab_instance):
        self._gitlab = gitlab_instance
        self._module = module
        self.project = project

    """
    @param milestone_id Title of the milestone
    """

    def get_milestone(self, milestone_id, group):
        milestones = []
        try:
            milestones = group.milestones.list(search=milestone_id)
        except gitlab.exceptions.GitlabGetError as e:
            self._module.fail_json(msg=f"Failed to list the Milestones: {e}")

        if len(milestones) > 1:
            self._module.fail_json(msg="Multiple Milestones matched search criteria.")
        if len(milestones) < 1:
            self._module.fail_json(msg="No Milestones matched search criteria.")
        if len(milestones) == 1:
            try:
                return group.milestones.get(id=milestones[0].id)
            except gitlab.exceptions.GitlabGetError as e:
                self._module.fail_json(msg=f"Failed to get the Milestones: {e}")

    """
    @param title Title of the Issue
    @param state_filter Issue's state to filter on
    """

    def get_issue(self, title, state_filter):
        issues = []
        try:
            issues = self.project.issues.list(query_parameters={"search": title, "in": "title", "state": state_filter})
        except gitlab.exceptions.GitlabGetError as e:
            self._module.fail_json(msg=f"Failed to list the Issues: {e}")

        if len(issues) > 1:
            self._module.fail_json(msg="Multiple Issues matched search criteria.")
        if len(issues) == 1:
            try:
                return self.project.issues.get(id=issues[0].iid)
            except gitlab.exceptions.GitlabGetError as e:
                self._module.fail_json(msg=f"Failed to get the Issue: {e}")

    """
    @param username Name of the user
    """

    def get_user(self, username):
        users = []
        try:
            users = [user for user in self.project.users.list(username=username, all=True) if user.username == username]
        except gitlab.exceptions.GitlabGetError as e:
            self._module.fail_json(msg=f"Failed to list the users: {e}")

        if len(users) > 1:
            self._module.fail_json(msg="Multiple Users matched search criteria.")
        elif len(users) < 1:
            self._module.fail_json(msg="No User matched search criteria.")
        else:
            return users[0]

    """
    @param users List of usernames
    """

    def get_user_ids(self, users):
        return [self.get_user(user).id for user in users]

    """
    @param options Options of the Issue
    """

    def create_issue(self, options):
        if self._module.check_mode:
            self._module.exit_json(changed=True, msg=f"Successfully created Issue '{options['title']}'.")

        try:
            return self.project.issues.create(options)
        except gitlab.exceptions.GitlabCreateError as e:
            self._module.fail_json(msg=f"Failed to create Issue: {e}")

    """
    @param issue Issue object to delete
    """

    def delete_issue(self, issue):
        if self._module.check_mode:
            self._module.exit_json(changed=True, msg=f"Successfully deleted Issue '{issue['title']}'.")

        try:
            return issue.delete()
        except gitlab.exceptions.GitlabDeleteError as e:
            self._module.fail_json(msg=f"Failed to delete Issue: '{e}'.")

    """
    @param issue Issue object to update
    @param options Options of the Issue
    """

    def update_issue(self, issue, options):
        if self._module.check_mode:
            self._module.exit_json(changed=True, msg=f"Successfully updated Issue '{issue['title']}'.")

        try:
            return self.project.issues.update(issue.iid, options)
        except gitlab.exceptions.GitlabUpdateError as e:
            self._module.fail_json(msg=f"Failed to update Issue {e}.")

    """
    @param issue Issue object to evaluate
    @param options New options to update Issue with
    """

    def issue_has_changed(self, issue, options):
        for key, value in options.items():
            if value is not None:
                if key == "milestone_id":
                    old_milestone = issue.milestone["id"] if issue.milestone else ""
                    if value != old_milestone:
                        return True
                elif key == "assignee_ids":
                    if value != sorted([user["id"] for user in issue.assignees]):
                        return True

                elif key == "labels":
                    if value != sorted(getattr(issue, key)):
                        return True

                elif getattr(issue, key) != value:
                    return True

        return False


def main():
    argument_spec = basic_auth_argument_spec()
    argument_spec.update(auth_argument_spec())
    argument_spec.update(
        assignee_ids=dict(type="list", elements="str"),
        description=dict(type="str"),
        description_path=dict(type="path"),
        issue_type=dict(type="str", default="issue", choices=["issue", "incident", "test_case"]),
        labels=dict(type="list", elements="str"),
        milestone_search=dict(type="str"),
        milestone_group_id=dict(type="str"),
        project=dict(type="str", required=True),
        state=dict(type="str", default="present", choices=["absent", "present"]),
        state_filter=dict(type="str", default="opened", choices=["opened", "closed"]),
        title=dict(type="str", required=True),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            ["api_username", "api_token"],
            ["api_username", "api_oauth_token"],
            ["api_username", "api_job_token"],
            ["api_token", "api_oauth_token"],
            ["api_token", "api_job_token"],
            ["description", "description_path"],
        ],
        required_together=[
            ["api_username", "api_password"],
            ["milestone_search", "milestone_group_id"],
        ],
        required_one_of=[["api_username", "api_token", "api_oauth_token", "api_job_token"]],
        supports_check_mode=True,
    )

    assignee_ids = module.params["assignee_ids"]
    description = module.params["description"]
    description_path = module.params["description_path"]
    issue_type = module.params["issue_type"]
    labels = module.params["labels"]
    milestone_id = module.params["milestone_search"]
    milestone_group_id = module.params["milestone_group_id"]
    project = module.params["project"]
    state = module.params["state"]
    state_filter = module.params["state_filter"]
    title = module.params["title"]

    # check prerequisites and connect to gitlab server
    gitlab_instance = gitlab_authentication(module, min_version="2.3.0")

    this_project = find_project(gitlab_instance, project)
    if this_project is None:
        module.fail_json(msg=f"Failed to get the project: {project}")

    this_gitlab = GitlabIssue(module=module, project=this_project, gitlab_instance=gitlab_instance)

    if milestone_id and milestone_group_id:
        this_group = find_group(gitlab_instance, milestone_group_id)
        if this_group is None:
            module.fail_json(msg=f"Failed to get the group: {milestone_group_id}")

        milestone_id = this_gitlab.get_milestone(milestone_id, this_group).id

    this_issue = this_gitlab.get_issue(title, state_filter)

    if state == "present":
        if description_path:
            try:
                with open(description_path, "rb") as f:
                    description = to_text(f.read(), errors="surrogate_or_strict")
            except OSError as e:
                module.fail_json(msg=f"Cannot open {description_path}: {e}")

        # sorting necessary in order to properly detect changes, as we don't want to get false positive
        # results due to differences in ids ordering;
        assignee_ids = sorted(this_gitlab.get_user_ids(assignee_ids)) if assignee_ids else assignee_ids
        labels = sorted(labels) if labels else labels

        options = {
            "title": title,
            "description": description,
            "labels": labels,
            "issue_type": issue_type,
            "milestone_id": milestone_id,
            "assignee_ids": assignee_ids,
        }

        if not this_issue:
            issue = this_gitlab.create_issue(options)
            module.exit_json(changed=True, msg=f"Created Issue '{title}'.", issue=issue.asdict())
        else:
            if this_gitlab.issue_has_changed(this_issue, options):
                issue = this_gitlab.update_issue(this_issue, options)
                module.exit_json(changed=True, msg=f"Updated Issue '{title}'.", issue=issue)
            else:
                module.exit_json(changed=False, msg=f"Issue '{title}' already exists", issue=this_issue.asdict())
    elif state == "absent":
        if not this_issue:
            module.exit_json(changed=False, msg=f"Issue '{title}' does not exist or has already been deleted.")
        else:
            issue = this_gitlab.delete_issue(this_issue)
            module.exit_json(changed=True, msg=f"Issue '{title}' deleted.", issue=issue)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2025 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the
# ansible.content_builder.
#
# Manually editing this file is not advised.
#
# To update the argspec make the desired changes
# in the documentation in the module file and re-run
# ansible.content_builder commenting out
# the path to external 'docstring' in build.yaml.
#
##############################################

"""
The arg spec for the ios_l3_interfaces module
"""


class L3_interfacesArgs(object):  # pylint: disable=R0903
    """The arg spec for the ios_l3_interfaces module"""

    argument_spec = {
        "config": {
            "type": "list",
            "elements": "dict",
            "options": {
                "name": {"type": "str", "required": True},
                "redirects": {"type": "bool"},
                "unreachables": {"type": "bool"},
                "ipv6_redirects": {"type": "bool"},
                "ipv6_unreachables": {"type": "bool"},
                "autostate": {"type": "bool"},
                "mac_address": {"type": "str"},
                "helper_addresses": {
                    "type": "dict",
                    "options": {
                        "ipv4": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "destination_ip": {"type": "str"},
                                "global": {"type": "bool"},
                                "vrf": {"type": "str"},
                            },
                        },
                    },
                },
                "ipv4": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "address": {"type": "str"},
                        "secondary": {"type": "bool"},
                        "dhcp_client": {"type": "str"},
                        "dhcp_hostname": {"type": "str"},
                        "dhcp": {
                            "type": "dict",
                            "options": {
                                "enable": {"type": "bool"},
                                "client_id": {"type": "str"},
                                "hostname": {"type": "str"},
                            },
                        },
                        "proxy_arp": {"type": "bool"},
                        "pool": {"type": "str"},
                        "redirects": {"type": "bool"},
                        "unreachables": {"type": "bool"},
                        "mtu": {"type": "int"},
                        "source_interface": {
                            "type": "dict",
                            "options": {
                                "name": {"type": "str"},
                                "poll": {"type": "bool"},
                                "point_to_point": {"type": "bool"},
                            },
                        },
                    },
                },
                "ipv6": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "address": {"type": "str"},
                        "autoconfig": {
                            "type": "dict",
                            "options": {
                                "enable": {"type": "bool"},
                                "default": {"type": "bool"},
                            },
                        },
                        "dhcp": {
                            "type": "dict",
                            "options": {
                                "enable": {"type": "bool"},
                                "rapid_commit": {"type": "bool"},
                            },
                        },
                        "enable": {"type": "bool"},
                        "anycast": {"type": "bool"},
                        "cga": {"type": "bool"},
                        "eui": {"type": "bool"},
                        "link_local": {"type": "bool"},
                        "segment_routing": {
                            "type": "dict",
                            "options": {
                                "enable": {"type": "bool"},
                                "default": {"type": "bool"},
                                "ipv6_sr": {"type": "bool"},
                            },
                        },
                    },
                },
            },
        },
        "running_config": {"type": "str"},
        "state": {
            "choices": [
                "merged",
                "replaced",
                "overridden",
                "deleted",
                "rendered",
                "gathered",
                "parsed",
            ],
            "default": "merged",
            "type": "str",
        },
    }  # pylint: disable=C0301
